
from registration.models import RegistrationProfile
from registration.forms import RegistrationForm

class KegbotRegistrationForm(RegistrationForm):
    def save(self, profile_callback=None):
        new_user = RegistrationProfile.objects.create_inactive_user(username=self.cleaned_data['username'],
                                                                    password=self.cleaned_data['password1'],
                                                                    email=self.cleaned_data['email'],
                                                                    send_email=False,
                                                                    profile_callback=profile_callback)
        new_user.is_active = True
        new_user.save()
        return new_user



# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Tasks for the Kegbot core."""

from kegbot.util import util
from pykeg.plugin import util as plugin_util
from pykeg import notification

from celery.decorators import task

def schedule_tasks(events):
    """Synchronously schedules tasks related to the given events."""
    for event in events:
        for plugin in plugin_util.get_plugins():
            plugin.handle_new_event(event)
    notification.handle_new_system_events(events)


@task
def handle_new_picture(picture_id):
    pass  # TODO(mikey): plugin support

@task
def ping():
    return True



# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Miscellaneous utility functions."""

# Note: imports should be limited to python stdlib, since methods here
# may be used in models.py, settings.py, etc.

import pkgutil
import os
import pkg_resources
import sys
import random

from django.core.exceptions import ImproperlyConfigured

def get_version():
    try:
        return pkg_resources.get_distribution('kegbot').version
    except pkg_resources.DistributionNotFound:
        return 'unknown'

def make_serial():
    '''Returns a random serial number.'''
    return '%016x' % random.randrange(0, 2**64 - 1)

def get_plugin_template_dirs(plugin_list):
    from django.utils import six
    if not six.PY3:
        fs_encoding = sys.getfilesystemencoding() or sys.getdefaultencoding()

    ret = []
    for plugin in plugin_list:
        plugin_module = '.'.join(plugin.split('.')[:-1])
        pkg = pkgutil.get_loader(plugin_module)
        if not pkg:
            raise ImproperlyConfigured('Cannot find plugin "%s"' % plugin)
        template_dir = os.path.join(os.path.dirname(pkg.filename), 'templates')
        if os.path.isdir(template_dir):
            if not six.PY3:
                template_dir = template_dir.decode(fs_encoding)
            ret.append(template_dir)
    return ret



from django.http import Http404
from django.conf import settings
from django.template.loader import render_to_string

from registration.models import RegistrationProfile
from registration.views import RegistrationView as BaseRegistrationView
from registration.backends.simple.views import RegistrationView as SimpleRegistrationView
from registration.backends.default.views import RegistrationView as DefaultRegistrationView

from pykeg.core import models

def send_activation_email(self, site):
    """
    Monkey-patched version of RegistrationProfile method, which
    adds Kegbot-specific context variables.
    """
    s = models.SiteSettings.get()
    site_name = s.title
    base_url = s.base_url()

    print 'send_activation_email', self.activation_key

    ctx_dict = {'activation_key': self.activation_key,
                'expiration_days': settings.ACCOUNT_ACTIVATION_DAYS,
                'site': site,
                'site_name': site_name,
                'base_url': base_url}
    subject = render_to_string('registration/activation_email_subject.txt',
                               ctx_dict)
    # Email subject *must not* contain newlines
    subject = ''.join(subject.splitlines())
    message = render_to_string('registration/activation_email.txt',
                               ctx_dict)
    self.user.email_user(subject, message, settings.DEFAULT_FROM_EMAIL)

RegistrationProfile.send_activation_email = send_activation_email


class RegistrationView(BaseRegistrationView):
    def __init__(self):
        self._simple = SimpleRegistrationView()
        self._default = DefaultRegistrationView()

    def _get_view(self, request):
        settings = request.kbsite.settings
        if not settings.registration_allowed:
            raise Http404('Registration is disabled.')
        if settings.registration_confirmation:
            return self._default
        else:
            return self._simple

    def get_success_url(self, request, user):
        settings = request.kbsite.settings
        if settings.registration_confirmation:
            return ('registration_complete', (), {})
        else:
            return 'kb-account-main'

    def registration_allowed(self, request):
        settings = request.kbsite.settings
        return settings.registration_allowed

    def register(self, request, **cleaned_data):
        return self._get_view(request).register(request, **cleaned_data)



# Copyright 2003-2009 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import uuid

from kegbot.util import units

from . import models

DEFAULT_KEG_SIZE_GALLONS = 15.5
DEFAULT_KEG_SIZE_DESCRIPTION = "Full Keg (half barrel)"

METER_NAME_0 = 'kegboard.flow0'
METER_NAME_1 = 'kegboard.flow1'

def get_default_keg_size():
    volume = units.Quantity(DEFAULT_KEG_SIZE_GALLONS, units.UNITS.USGallon)
    volume_int = volume.Amount(in_units=units.RECORD_UNIT)
    return models.KegSize.objects.get_or_create(name=DEFAULT_KEG_SIZE_DESCRIPTION,
        volume_ml=volume_int)[0]

def db_is_installed():
    return models.KegbotSite.objects.all().count() > 0

class AlreadyInstalledError(Exception):
    """Thrown when database is already installed."""

def set_defaults(force=False, set_is_setup=False):
    """Creates a new site and sets defaults, returning that site."""
    if not force and db_is_installed():
        raise AlreadyInstalledError("Database is already installed.")

    site = models.KegbotSite.get()
    if set_is_setup and not site.is_setup:
        site.is_setup = True
        site.save()

    # KegTap defaults
    main_tap = models.KegTap(name='Main Tap', meter_name=METER_NAME_0)
    main_tap.save()
    secondary_tap = models.KegTap(name='Second Tap', meter_name=METER_NAME_1)
    secondary_tap.save()

    # brewer defaults
    unk_brewer = models.Brewer(name='Unknown Brewer')
    unk_brewer.save()

    # beerstyle defaults
    unk_style = models.BeerStyle(name='Unknown Style')
    unk_style.save()

    # beertype defaults
    unk_type = models.BeerType(name="Unknown Beer", brewer=unk_brewer, style=unk_style)
    unk_type.save()

    # KegSize defaults - from http://en.wikipedia.org/wiki/Keg#Size
    default_sizes = (
      (DEFAULT_KEG_SIZE_GALLONS, DEFAULT_KEG_SIZE_DESCRIPTION),
      (13.2, "Import Keg (European barrel)"),
      (7.75, "Pony Keg (quarter barrel)"),
      (6.6, "European Half Barrel"),
      (5.23, "Sixth Barrel (torpedo keg)"),
      (5.0, "Corny Keg"),
      (1.0, "Mini Keg"),
    )
    for gallons, description in default_sizes:
        volume = units.Quantity(gallons, units.UNITS.USGallon)
        volume_int = volume.Amount(in_units=units.RECORD_UNIT)

        ks = models.KegSize(
          name=description,
          volume_ml=volume_int,
        )
        ks.save()
    return site



# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Routines from converting data to and from Protocol Buffer format."""

import pytz

from django.conf import settings

from kegbot.api import api_pb2
from kegbot.api import models_pb2
from kegbot.api import protoutil
from kegbot.util import util

from pykeg.contrib.soundserver import models as soundserver_models
from pykeg.core import models

_CONVERSION_MAP = {}

def converts(kind):
    def decorate(f):
        global _CONVERSION_MAP
        _CONVERSION_MAP[kind] = f
        return f
    return decorate

def datestr(dt):
    if settings.USE_TZ:
        return dt.isoformat()
    try:
        # Convert from local to UTC.
        # TODO(mikey): handle incoming datetimes with tzinfo.
        dt = util.local_to_utc(dt, settings.TIME_ZONE)
    except pytz.UnknownTimeZoneError:
        pass
    return dt.strftime('%Y-%m-%dT%H:%M:%SZ')

def ToProto(obj, full=False):
    """Converts the object to protocol format."""
    if obj is None:
        return None
    kind = obj.__class__
    if hasattr(obj, '__iter__'):
        return [ToProto(item, full) for item in obj]
    elif kind in _CONVERSION_MAP:
        return _CONVERSION_MAP[kind](obj, full)
    else:
        raise ValueError, "Unknown object type: %s" % kind

def ToDict(obj, full=False):
    res = ToProto(obj, full)
    if hasattr(res, '__iter__'):
        return [protoutil.ProtoMessageToDict(m) for m in res]
    else:
        return protoutil.ProtoMessageToDict(res)

### Model conversions

@converts(models.AuthenticationToken)
def AuthTokenToProto(record, full=False):
    ret = models_pb2.AuthenticationToken()
    ret.id = record.id
    ret.auth_device = record.auth_device
    ret.token_value = record.token_value
    if record.user_id:
        ret.username = str(record.user.username)
        ret.user.MergeFrom(ToProto(record.user))
    if record.nice_name:
        ret.nice_name = record.nice_name
    ret.created_time = datestr(record.created_time)
    ret.enabled = record.enabled
    if record.expire_time:
        ret.expire_time = datestr(record.expire_time)
    if record.pin:
        ret.pin = record.pin
    return ret

@converts(models.Picture)
def PictureToProto(record, full=False):
    ret = models_pb2.Image()
    ret.url = record.resized.url
    ret.original_url = record.image.url
    ret.thumbnail_url = record.thumbnail.url
    # TODO(mikey): This can be expensive depending on the storage backend
    # (attempts to fetch image).
    #try:
    #  ret.width = record.image.width
    #  ret.height = record.image.height
    #except IOError:
    #  pass
    return ret

@converts(models.PourPicture)
def PourPictureToProto(record, full=False):
    ret = PictureToProto(record.picture)
    if record.time:
        ret.time = datestr(record.time)
    if record.caption:
        ret.caption = record.caption
    if record.user_id:
        ret.user_id = record.user.username
    if record.keg_id:
        ret.keg_id = record.keg_id
    if record.session_id:
        ret.session_id = record.session_id
    if record.drink_id:
        ret.drink_id = record.drink_id
    return ret

@converts(models.BeerStyle)
def BeerStyleToProto(style, full=False):
    ret = models_pb2.BeerStyle()
    ret.id = style.id
    ret.name = style.name
    return ret

@converts(models.BeerType)
def BeerTypeToProto(beertype, full=False):
    ret = models_pb2.BeerType()
    ret.id = str(beertype.id)
    ret.name = beertype.name
    ret.brewer_id = str(beertype.brewer_id)
    ret.style_id = str(beertype.style_id)
    if beertype.edition is not None:
        ret.edition = beertype.edition
    # TODO(mikey): guarantee this at DB level
    abv = beertype.abv or 0.0
    ret.abv = max(min(abv, 100.0), 0.0)
    if beertype.calories_oz is not None:
        ret.calories_oz = beertype.calories_oz
    if beertype.carbs_oz is not None:
        ret.carbs_oz = beertype.carbs_oz
    if beertype.specific_gravity is not None:
        ret.specific_gravity = beertype.specific_gravity
    if beertype.original_gravity is not None:
        ret.original_gravity = beertype.original_gravity
    if beertype.image:
        ret.image.MergeFrom(ToProto(beertype.image))
    return ret

@converts(models.Brewer)
def BrewerToProto(brewer, full=False):
    ret = models_pb2.Brewer()
    ret.id = str(brewer.id)
    ret.name = brewer.name
    if brewer.country is not None:
        ret.country = brewer.country
    if brewer.origin_state is not None:
        ret.origin_state = brewer.origin_state
    if brewer.origin_city is not None:
        ret.origin_city = brewer.origin_city
    if brewer.production is not None:
        ret.production = brewer.production
    if brewer.url is not None:
        ret.url = brewer.url
    if brewer.description is not None:
        ret.description = brewer.description
    if brewer.image:
        ret.image.MergeFrom(ToProto(brewer.image))
    return ret

@converts(models.Drink)
def DrinkToProto(drink, full=False):
    ret = models_pb2.Drink()
    ret.id = drink.id
    ret.url = drink.get_absolute_url()
    ret.ticks = drink.ticks
    ret.volume_ml = drink.volume_ml
    ret.session_id = drink.session_id
    ret.time = datestr(drink.time)
    ret.duration = drink.duration
    if drink.keg:
        ret.keg_id = drink.keg_id
    if drink.user_id:
        ret.user_id = drink.user.username
    if drink.shout:
        ret.shout = drink.shout
    if drink.tick_time_series:
        ret.tick_time_series = drink.tick_time_series

    if full:
        if drink.user_id:
            ret.user.MergeFrom(ToProto(drink.user))
        if drink.keg:
            ret.keg.MergeFrom(ToProto(drink.keg))
        if drink.session:
            ret.session.MergeFrom(ToProto(drink.session))
        for i in drink.pictures.all():
            ret.images.add().MergeFrom(ToProto(i))
    return ret

@converts(models.Keg)
def KegToProto(keg, full=False):
    ret = models_pb2.Keg()
    ret.id = keg.id
    ret.url = keg.get_absolute_url()
    ret.type_id = str(keg.type_id)
    ret.size_id = keg.size_id
    ret.size_name = keg.size.name
    ret.size_volume_ml = keg.size.volume_ml
    rem = float(keg.remaining_volume())
    ret.volume_ml_remain = rem
    ret.percent_full = keg.percent_full()
    ret.start_time = datestr(keg.start_time)
    ret.end_time = datestr(keg.end_time)
    if hasattr(ret, 'online'):
        ret.online = keg.online
    if keg.description is not None:
        ret.description = keg.description
    ret.spilled_ml = keg.spilled_ml

    if full:
        if keg.type:
            ret.type.MergeFrom(ToProto(keg.type))
        if keg.size:
            ret.size.MergeFrom(ToProto(keg.size))

    return ret

@converts(models.KegSize)
def KegSizeToProto(size, full=False):
    ret = models_pb2.KegSize()
    ret.id = size.id
    ret.name = size.name
    ret.volume_ml = size.volume_ml
    return ret

@converts(models.KegTap)
def KegTapToProto(tap, full=False):
    ret = models_pb2.KegTap()
    ret.id = tap.id
    ret.name = tap.name
    ret.meter_name = tap.meter_name
    ret.relay_name = tap.relay_name or ''
    ret.ml_per_tick = tap.ml_per_tick
    if tap.description is not None:
        ret.description = tap.description
    if tap.current_keg:
        ret.current_keg_id = tap.current_keg_id
        if full:
            ret.current_keg.MergeFrom(ToProto(tap.current_keg, full=True))

    if tap.temperature_sensor:
        ret.thermo_sensor_id = tap.temperature_sensor_id
        log = tap.temperature_sensor.LastLog()
        if log:
            ret.last_temperature.MergeFrom(ToProto(log))
    return ret

@converts(models.DrinkingSession)
def SessionToProto(record, full=False):
    ret = models_pb2.Session()
    ret.id = record.id
    ret.url = record.get_absolute_url()
    ret.start_time = datestr(record.start_time)
    ret.end_time = datestr(record.end_time)
    ret.volume_ml = record.volume_ml
    ret.name = record.name or ''

    if full:
        #ret.stats.MergeFrom(record.GetStats())
        ret.is_active = record.IsActiveNow()
    return ret

@converts(models.Thermolog)
def ThermoLogToProto(record, full=False):
    ret = models_pb2.ThermoLog()
    ret.id = record.id
    ret.sensor_id = record.sensor_id
    ret.temperature_c = record.temp
    ret.time = datestr(record.time)
    return ret

@converts(models.ThermoSensor)
def ThermoSensorToProto(record, full=False):
    ret = models_pb2.ThermoSensor()
    ret.id = record.id
    ret.sensor_name = record.raw_name
    ret.nice_name = record.nice_name
    return ret

@converts(models.User)
def UserToProto(user, full=False):
    ret = models_pb2.User()
    ret.username = user.username
    ret.url = user.get_profile().get_absolute_url()
    ret.is_active = user.is_active
    if full:
        ret.first_name = user.first_name
        ret.last_name = user.last_name
        ret.email = user.email
        ret.is_staff = user.is_staff
        ret.is_active = user.is_active
        ret.is_superuser = user.is_superuser
        ret.last_login = datestr(user.last_login)
        ret.date_joined = datestr(user.date_joined)
    profile = user.get_profile()
    if profile.mugshot_id:
        ret.image.MergeFrom(ToProto(profile.mugshot))
    return ret

@converts(models.UserProfile)
def UserProfileToProto(record, full=False):
    ret = models_pb2.UserProfile()
    ret.username = record.user.username
    return ret

@converts(models.SystemStats)
@converts(models.UserStats)
@converts(models.KegStats)
@converts(models.SessionStats)
def SystemStatsToProto(record, full=False):
    return protoutil.DictToProtoMessage(record.stats, models_pb2.Stats())

@converts(models.SystemEvent)
def SystemEventToProto(record, full=False):
    ret = models_pb2.SystemEvent()
    ret.id = record.id
    ret.kind = record.kind
    ret.time = datestr(record.time)

    if record.drink_id:
        ret.drink_id = record.drink_id
        if full:
            ret.drink.MergeFrom(ToProto(record.drink, full=True))
    if record.keg_id:
        ret.keg_id = record.keg_id
        if full:
            ret.keg.MergeFrom(ToProto(record.keg, full=True))
    if record.session_id:
        ret.session_id = record.session_id
        if full:
            ret.session.MergeFrom(ToProto(record.session, full=True))
    if record.user_id:
        ret.user_id = str(record.user.username)
        if full:
            ret.user.MergeFrom(ToProto(record.user, full=True))

    image = None
    if record.kind in ('drink_poured', 'session_started', 'session_joined') and record.user:
        image = record.user.get_profile().mugshot
    elif record.kind in ('keg_tapped', 'keg_ended'):
        if record.keg.type and record.keg.type.image:
            image = record.keg.type.image
    if image:
        ret.image.MergeFrom(ToProto(image))

    return ret

@converts(soundserver_models.SoundEvent)
def SoundEventToProto(record, full=False):
    ret = models_pb2.SoundEvent()
    ret.event_name = record.event_name
    ret.event_predicate = record.event_predicate
    ret.sound_url = record.soundfile.sound.url
    if record.user_id:
        ret.user = record.user.username
    return ret

# Composite messages

def GetDrinkDetail(drink):
    ret = api_pb2.DrinkDetail()
    ret.drink.MergeFrom(ToProto(drink))
    if drink.user_id:
        ret.user.MergeFrom(ToProto(drink.user))
    if drink.keg_id:
        ret.keg.MergeFrom(ToProto(drink.keg))
    if drink.session_id:
        ret.session.MergeFrom(ToProto(drink.session))
    for i in drink.pictures.all().iterator():
        ret.images.add().MergeFrom(ToProto(i))
    return ret



# Copyright 2003-2011 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.test import TransactionTestCase

from kegbot.api import models_pb2
from kegbot.api.protoutil import ProtoMessageToDict
from kegbot.util import util

from . import backend
from . import models
from . import stats
from .testutils import make_datetime

class StatsTestCase(TransactionTestCase):
    reset_sequences = True

    def setUp(self):
        self.backend = backend.KegbotBackend()

        test_usernames = ('user1', 'user2', 'user3')
        self.users = [self.backend.create_new_user(name) for name in test_usernames]

        self.taps = [
            self.backend.create_tap('tap1', 'kegboard.flow0', ml_per_tick=1/2200.0),
            self.backend.create_tap('tap2', 'kegboard.flow1', ml_per_tick=1/2200.0),
        ]

        self.keg = self.backend.start_keg('kegboard.flow0', beer_name='Unknown',
            brewer_name='Unknown', style_name='Unknown')

    def testStuff(self):
        site = models.KegbotSite.get()
        stats = site.GetStats()
        self.assertEquals(stats, {})

        now = make_datetime(2012, 1, 2, 12, 00)
        self.maxDiff = None

        d = self.backend.record_drink('kegboard.flow0', ticks=1, volume_ml=100,
            username='user1', pour_time=now)
        expected = util.AttrDict({
            u'volume_by_year': {u'2012': 100.0},
            u'total_pours': 1,
            u'has_guest_pour': False,
            u'greatest_volume_ml': 100.0,
            u'registered_drinkers': [u'user1'],
            u'volume_by_day_of_week': {u'1': 100.0},
            u'greatest_volume_id': d.id,
            u'volume_by_drinker': {u'user1': 100.0},
            u'last_drink_id': d.id,
            u'sessions_count': 1,
            u'average_volume_ml': 100.0,
            u'total_volume_ml': 100.0
        })
        stats = site.GetStats()
        self.assertDictEqual(expected, stats)

        now = make_datetime(2012, 1, 3, 12, 00)
        d = self.backend.record_drink('kegboard.flow0', ticks=200,
            volume_ml=200, username='user2', pour_time=now)
        stats = site.GetStats()
        expected.total_pours = 2
        expected.greatest_volume_ml = 200.0
        expected.greatest_volume_id = d.id
        expected.registered_drinkers.append(u'user2')
        expected.volume_by_drinker[u'user2'] = 200.0
        expected.last_drink_id = d.id
        expected.average_volume_ml = 150.0
        expected.total_volume_ml = 300.0
        expected.volume_by_day_of_week[u'2'] = 200.0
        expected.volume_by_year[u'2012'] = 300.0
        expected.sessions_count = 2
        print 'EXPECTED'
        import pprint
        pprint.pprint(expected)
        print '----'
        print 'ACTUAL'
        pprint.pprint(stats)
        self.assertDictEqual(expected, stats)



# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Kegweb RESTful API views."""

import datetime
import logging
from functools import wraps

from django.contrib.auth import login as auth_login
from django.contrib.auth import logout as auth_logout
from django.contrib.auth.forms import AuthenticationForm
from django.db.utils import IntegrityError
from django.utils import timezone

from django.http import Http404
from django.shortcuts import get_object_or_404
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods

from kegbot.api import kbapi

from pykeg.contrib.soundserver import models as soundserver_models
from pykeg.core import backend
from pykeg.core import models
from pykeg.core import util as core_util
from pykeg.proto import protolib
from pykeg.web.api import forms
from pykeg.web.api import util
from pykeg.web.kegadmin.forms import ChangeKegForm

from pykeg.web import tasks

_LOGGER = logging.getLogger(__name__)

### Decorators

def auth_required(view_func):
    def wrapped_view(*args, **kwargs):
        return view_func(*args, **kwargs)
    util.set_needs_auth(wrapped_view)
    return wraps(view_func)(wrapped_view)

### Helpers

def _form_errors(form):
    ret = {}
    for field in form:
        if field.errors:
            name = field.html_name
            ret[name] = []
            for error in field.errors:
                ret[name].append(error)
    return ret

### Endpoints

def all_kegs(request):
    return models.Keg.objects.all().order_by('-start_time')

def all_drinks(request, limit=100):
    qs = models.Drink.objects.all()
    if 'start' in request.GET:
        try:
            start = int(request.GET['start'])
            qs = qs.filter(id__lte=start)
        except ValueError:
            pass
    qs = qs.order_by('-id')
    qs = qs[:limit]
    start = qs[0].id
    return qs

def get_drink(request, drink_id):
    drink = get_object_or_404(models.Drink, id=drink_id)
    return protolib.ToProto(drink, full=True)

@csrf_exempt
@auth_required
def add_drink_photo(request, drink_id):
    if request.method != 'POST':
        raise Http404('Method not supported')
    drink = get_object_or_404(models.Drink, id=drink_id)
    pour_pic = _save_pour_pic(request, drink)
    return protolib.ToProto(pour_pic, full=True)

def _save_pour_pic(request, drink):
    pic = models.Picture.objects.create(image=request.FILES['photo'])
    pour_pic = models.PourPicture.objects.create(picture_id=pic.id,
        drink=drink,
        user=drink.user,
        keg=drink.keg,
        session=drink.session)
    tasks.handle_new_picture.delay(pour_pic.id)
    return pour_pic

def get_session(request, session_id):
    session = get_object_or_404(models.DrinkingSession, id=session_id)
    return protolib.ToProto(session, full=True)

def get_session_stats(request, session_id):
    session = get_object_or_404(models.DrinkingSession, id=session_id)
    return session.GetStats()

def get_keg(request, keg_id):
    keg = get_object_or_404(models.Keg, id=keg_id)
    return protolib.ToProto(keg, full=True)

def get_keg_drinks(request, keg_id):
    keg = get_object_or_404(models.Keg, id=keg_id)
    return keg.drinks.all()

def get_keg_events(request, keg_id):
    keg = get_object_or_404(models.Keg, id=keg_id)
    events = keg.events.all()
    events = apply_since(request, events)
    return events

def get_keg_sizes(request):
    return models.KegSize.objects.all()

@require_http_methods(["POST"])
@auth_required
@csrf_exempt
def end_keg(request, keg_id):
    keg = get_object_or_404(models.Keg, id=keg_id)
    tap = keg.current_tap
    keg = request.backend.end_keg(tap)
    return protolib.ToProto(keg, full=True)

def all_sessions(request):
    return models.DrinkingSession.objects.all()

def current_session(request):
    try:
        latest = models.DrinkingSession.objects.latest()
        if latest.IsActiveNow():
            return latest
    except models.DrinkingSession.DoesNotExist:
        raise Http404

def all_events(request):
    events = models.SystemEvent.objects.all().order_by('-id')
    events = apply_since(request, events)
    events = events[:10]
    return [protolib.ToProto(e, full=True) for e in events]

def apply_since(request, query):
    """Restricts the query to `since` events, if given."""
    since_str = request.GET.get('since')
    if since_str:
        try:
            since = int(since_str)
            return query.filter(id__gt=since)
        except (ValueError, TypeError):
            pass
    return query

@auth_required
def all_sound_events(request):
    return soundserver_models.SoundEvent.objects.all()

def get_keg_sessions(request, keg_id):
    keg = get_object_or_404(models.Keg, id=keg_id)
    sessions = [c.session for c in keg.keg_session_chunks.all()]
    return sessions

def get_keg_stats(request, keg_id):
    keg = get_object_or_404(models.Keg, id=keg_id)
    return keg.GetStats()

def get_system_stats(request):
    return models.KegbotSite.get().GetStats()

def get_info(request):
    return {'kegbot_server_version': core_util.get_version()}

def all_taps(request):
    return models.KegTap.objects.all().order_by('name')

@auth_required
def user_list(request):
    return models.User.objects.filter(is_active=True).order_by('username')

def get_user(request, username):
    user = get_object_or_404(models.User, username=username)
    return protolib.ToProto(user, full=True)

def get_user_drinks(request, username):
    user = get_object_or_404(models.User, username=username)
    return user.drinks.all()

def get_user_events(request, username):
    user = get_object_or_404(models.User, username=username)
    return user.events.all()

def get_user_stats(request, username):
    user = get_object_or_404(models.User, username=username)
    return user.get_profile().GetStats()

@auth_required
def get_auth_token(request, auth_device, token_value):
    tok = request.backend.get_auth_token(auth_device, token_value)
    return tok

@csrf_exempt
@auth_required
def assign_auth_token(request, auth_device, token_value):
    if not request.POST:
        raise kbapi.BadRequestError('POST required.')

    form = forms.AssignTokenForm(request.POST)
    if not form.is_valid():
        errors = _form_errors(form)
        raise kbapi.BadRequestError(errors)

    b = request.backend
    username = form.cleaned_data['username']

    user = backend.get_user(username)
    if not user:
        raise kbapi.BadRequestError("User does not exist")

    try:
        tok = b.get_auth_token(auth_device, token_value)
    except backend.NoTokenError:
        tok = b.create_auth_token(auth_device, token_value, username=username)

    if tok.user != user:
        if tok.user:
            raise kbapi.BadRequestError("Token is already bound to a user")
        tok.user = user
        tok.save()
    return tok

def all_thermo_sensors(request):
    return models.ThermoSensor.objects.all()

def _get_sensor_or_404(request, sensor_name):
    try:
        sensor = models.ThermoSensor.objects.get(raw_name=sensor_name)
    except models.ThermoSensor.DoesNotExist:
        try:
            sensor = models.ThermoSensor.objects.get(nice_name=sensor_name)
        except models.ThermoSensor.DoesNotExist:
            raise Http404
    return sensor

@csrf_exempt
def get_thermo_sensor(request, sensor_name):
    if request.method == 'POST':
        return _thermo_sensor_post(request, sensor_name)
    else:
        return _thermo_sensor_get(request, sensor_name)

def _thermo_sensor_get(request, sensor_name):
    sensor = _get_sensor_or_404(request, sensor_name)
    logs = sensor.thermolog_set.all()
    if not logs:
        last_temp = None
        last_time = None
    else:
        last_temp = logs[0].temp
        last_time = logs[0].time
    res = {
      'sensor': sensor,
      'last_temp': last_temp,
      'last_time': last_time,
    }
    return res

@auth_required
def _thermo_sensor_post(request, sensor_name):
    form = forms.ThermoPostForm(request.POST)
    if not form.is_valid():
        raise kbapi.BadRequestError, _form_errors(form)
    cd = form.cleaned_data
    sensor, created = models.ThermoSensor.objects.get_or_create(raw_name=sensor_name)
    # TODO(mikey): use form fields to compute `when`
    return request.backend.log_sensor_reading(sensor.raw_name, cd['temp_c'])

def get_thermo_sensor_logs(request, sensor_name):
    sensor = _get_sensor_or_404(request, sensor_name)
    return sensor.thermolog_set.all()[:60*2]

def get_api_key(request):
    user = request.user
    api_key = ''
    if user and (user.is_staff or user.is_superuser):
        api_key = user.get_profile().GetApiKey()
    return {'api_key': api_key}

@csrf_exempt
def tap_detail(request, tap_id):
    tap = get_object_or_404(models.KegTap, meter_name=tap_id)
    if request.method == 'POST':
        return _tap_detail_post(request, tap)
    elif request.method == 'GET':
        return _tap_detail_get(request, tap)
    else:
        raise kbapi.BadRequestError('Method not supported')

def _tap_detail_get(request, tap):
    return protolib.ToProto(tap, full=True)

@csrf_exempt
@auth_required
def tap_calibrate(request, tap_id):
    # TODO(mikey): This would make more semantic sense as PATCH /taps/tap-name/,
    # but Django's support for non-POST verbs is poor (specifically wrt request
    # body/form handling).
    tap = get_object_or_404(models.KegTap, meter_name=tap_id)
    form = forms.CalibrateTapForm(request.POST)
    if form.is_valid():
        tap.ml_per_tick = form.cleaned_data['ml_per_tick']
        tap.save()
    else:
        raise kbapi.BadRequestError, _form_errors(form)
    return protolib.ToProto(tap, full=True)

@csrf_exempt
@auth_required
def tap_spill(request, tap_id):
    tap = get_object_or_404(models.KegTap, meter_name=tap_id)
    if not tap.current_keg:
        raise kbapi.BadRequestError('No keg on tap.')
    form = forms.TapSpillForm(request.POST)
    if form.is_valid():
        tap.current_keg.spilled_ml += form.cleaned_data['volume_ml']
        tap.current_keg.save()
    else:
        raise kbapi.BadRequestError, _form_errors(form)
    return protolib.ToProto(tap, full=True)

@csrf_exempt
@auth_required
def tap_activate(request, tap_id):
    tap = get_object_or_404(models.KegTap, meter_name=tap_id)
    form = ChangeKegForm(request.POST)
    if form.is_valid():
        form.save(tap)
    else:
        raise kbapi.BadRequestError, _form_errors(form)
    return protolib.ToProto(tap, full=True)

@auth_required
def _tap_detail_post(request, tap):
    form = forms.DrinkPostForm(request.POST)
    if not form.is_valid():
        raise kbapi.BadRequestError, _form_errors(form)
    cd = form.cleaned_data
    if cd.get('pour_time') and cd.get('now'):
        pour_time = datetime.datetime.fromtimestamp(cd.get('pour_time'))
        pour_now = datetime.datetime.fromtimestamp(cd.get('now'))
        pour_time_ago = pour_now - pour_time
        pour_time = timezone.now() - pour_time_ago
    else:
        pour_time = None
    duration = cd.get('duration')
    if duration is None:
        duration = 0
    try:
        drink = request.backend.record_drink(tap_name=tap.meter_name,
          ticks=cd['ticks'],
          volume_ml=cd.get('volume_ml'),
          username=cd.get('username'),
          pour_time=pour_time,
          duration=duration,
          shout=cd.get('shout'),
          tick_time_series=cd.get('tick_time_series'))
        if 'photo' in request.FILES:
            _save_pour_pic(request, drink)
        return protolib.ToProto(drink, full=True)
    except backend.BackendError, e:
        raise kbapi.ServerError(str(e))

@csrf_exempt
@auth_required
def cancel_drink(request):
    if request.method != 'POST':
        raise kbapi.BadRequestError('POST required')
    form = forms.CancelDrinkForm(request.POST)
    if not form.is_valid():
        raise kbapi.BadRequestError, _form_errors(form)
    cd = form.cleaned_data
    try:
        res = request.backend.cancel_drink(drink_id=cd.get('id'), spilled=cd.get('spilled', False))
        return protolib.ToProto(res, full=True)
    except backend.BackendError, e:
        raise kbapi.ServerError(str(e))

@csrf_exempt
def login(request):
    if request.POST:
        form = AuthenticationForm(data=request.POST)
        if form.is_valid():
            auth_login(request, form.get_user())
            if request.session.test_cookie_worked():
                request.session.delete_test_cookie()
            return {'result': 'ok'}
        else:
            raise kbapi.PermissionDeniedError('Login failed.')
    raise kbapi.BadRequestError('POST required.')

def logout(request):
    auth_logout(request)
    return {'result': 'ok'}

@csrf_exempt
@auth_required
def register(request):
    if not request.POST:
        raise kbapi.BadRequestError('POST required.')
    form = forms.RegisterForm(request.POST)
    errors = {}
    if not form.is_valid():
        errors = _form_errors(form)
    else:
        username = form.cleaned_data['username']
        try:
            u = models.User()
            u.username = username
            u.email = form.cleaned_data['email']
            password = form.cleaned_data.get('password')
            if password:
                u.set_password(password)
            else:
                # Must set password using recovery process.
                u.set_unusable_password()
            u.save()
            if 'photo' in request.FILES:
                pic = models.Picture.objects.create()
                photo = request.FILES['photo']
                pic.image.save(photo.name, photo)
                pic.save()
                profile = u.get_profile()
                profile.mugshot = pic
                profile.save()
            return protolib.ToProto(u, full=True)
        except IntegrityError:
            user_errs = errors.get('username', [])
            user_errs.append('Username not available.')
            errors['username'] = user_errs
    raise kbapi.BadRequestError(errors)

@csrf_exempt
@auth_required
def user_photo(request, username):
    user = get_object_or_404(models.User, username=username)
    if request.method == 'POST':
        return post_user_photo(request, user)
    else:
        return get_user_photo(request, user)

def get_user_photo(request, user):
    mugshot = user.get_profile().mugshot
    if not mugshot:
        return {}
    return mugshot

def post_user_photo(request, user):
    photo_file = request.FILES.get('photo')
    if not photo_file:
        raise kbapi.BadRequestError('The file "photo" is required.')

    pic = models.Picture.objects.create()
    pic.image.save(photo_file.name, photo_file)
    pic.save()
    profile = user.get_profile()
    profile.mugshot = pic
    profile.save()
    return protolib.ToProto(pic)


def default_handler(request):
    raise Http404, "Not an API endpoint: %s" % request.path[:100]



from django.conf import settings

import pykeg
from pykeg.core import models
from pykeg.core import util
from pykeg.web.kegweb.forms import LoginForm

def kbsite(request):
    kbsite = getattr(request, 'kbsite', None)

    ret = {
      'DEBUG': settings.DEBUG,
      'EMBEDDED': getattr(settings, 'EMBEDDED', False),
      'EPOCH': pykeg.EPOCH,
      'VERSION': util.get_version(),
      'HAVE_SESSIONS': False,
      'HAVE_ADMIN': settings.KEGBOT_ENABLE_ADMIN,
      'GOOGLE_ANALYTICS_ID': None,
      'kbsite': kbsite,
      'request_path': request.path,
      'login_form': LoginForm(initial={'next_page': request.path}),
      'guest_info': {
        'name': 'guest',
        'image': None,
      },
      'PLUGINS': getattr(request, 'plugins', {}),
    }

    if kbsite:
        ret['guest_info']['name'] = kbsite.settings.guest_name
        ret['guest_info']['image'] = kbsite.settings.guest_image
        ret['SERIAL_NUMBER'] = kbsite.serial_number
        ret['HAVE_SESSIONS'] = models.DrinkingSession.objects.all().count() > 0
        ret['GOOGLE_ANALYTICS_ID'] = kbsite.settings.google_analytics_id

    return ret



# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Celery tasks for Webhook plugin."""

from celery.task import task
from pykeg.plugin import util
from pykeg.proto import protolib
from pykeg.core.util import get_version
from kegbot.util import kbjson
import requests
import urllib2

logger = util.get_logger(__name__)

@task(expires=180)
def post_webhook(url, event):
    """Posts an event to the supplied URL.

    The request body is a JSON dictionary of:
      * type: webhook message type (currently always 'event')
      * data: webhook data (the event payload)

    Event payloads are in the same format as the /api/events/ endpoint.
    """
    logger.info('Posting webhook: url=%s event=%s' % (url, event))

    event_dict = protolib.ToDict(event)
    hook_dict = {
        'type': 'event',
        'data': event_dict,
    }

    headers = {
        'content-type': 'application/json',
        'user-agent': 'Kegbot/%s' % get_version(),
    }
    
    try:
        return requests.post(url, data=kbjson.dumps(hook_dict), headers=headers)
    except requests.exceptions.RequestException, e:
        logger.warning('Error posting hook: %s' % e)
        return False


# Copyright 2014 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for notification module."""

from django.core import mail
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase
from django.test.utils import override_settings
from pykeg.core import models as core_models
from pykeg.core import backend
from pykeg.core import defaults

from pykeg import notification
from pykeg.notification import models
from pykeg.notification.backends.base import BaseNotificationBackend


class TestBackendA(BaseNotificationBackend):
    def notify(self, event, user):
        pass


class TestBackendB(BaseNotificationBackend):
    def notify(self, event, user):
        pass


class StandaloneNotificationTestCase(TestCase):
    """Standalone notification tests, no other components involved."""

    @override_settings(NOTIFICATION_BACKENDS=[
      'pykeg.notification.notification_test.TestBackendA',
      'pykeg.notification.notification_test.TestBackendB',
    ])
    def test_get_backends(self):
        backends = notification.get_backends()
        self.assertEquals(2, len(backends))
        self.assertIsInstance(backends[0], TestBackendA, 'Expected TestBackendA obj')
        self.assertIsInstance(backends[1], TestBackendB, 'Expected TestBackendB obj')

    @override_settings(NOTIFICATION_BACKENDS=['pykeg.bougus.NoBackend'])
    def test_improperly_configured(self):
        self.assertRaises(ImproperlyConfigured, notification.get_backends)


class NotificationTestCase(TestCase):
    def setUp(self):
        self.backend = backend.KegbotBackend()
        defaults.set_defaults(set_is_setup=True)

        self.user = core_models.User.objects.create(username='notification_user',
            email='test@example')

    @override_settings(NOTIFICATION_BACKENDS=['pykeg.notification.notification_test.CaptureBackend'])
    def test_notifications(self):
        class CaptureBackend(BaseNotificationBackend):
            """Notification backend which captures calls."""
            captured = []
            def notify(self, event, user):
                self.captured.append((event, user))

        SystemEvent = core_models.SystemEvent

        backends = [CaptureBackend()]
        captured = CaptureBackend.captured
        self.assertEquals(0, len(captured))

        prefs = models.NotificationSettings.objects.create(user=self.user,
            backend='pykeg.notification.notification_test.CaptureBackend',
            keg_tapped=False, session_started=False, keg_volume_low=False,
            keg_ended=False)

        event = SystemEvent(kind=SystemEvent.KEG_TAPPED)
        notification.handle_single_event(event, backends)
        self.assertEquals(0, len(captured))
        prefs.keg_tapped = True
        prefs.save()
        notification.handle_single_event(event, backends)
        self.assertEquals(1, len(captured))
        del captured[:]

        event = SystemEvent(kind=SystemEvent.SESSION_STARTED)
        notification.handle_single_event(event, backends)
        self.assertEquals(0, len(captured))
        prefs.session_started = True
        prefs.save()
        notification.handle_single_event(event, backends)
        self.assertEquals(1, len(captured))
        del captured[:]

        event = SystemEvent(kind=SystemEvent.KEG_ENDED)
        notification.handle_single_event(event, backends)
        self.assertEquals(0, len(captured))
        prefs.keg_ended = True
        prefs.save()
        notification.handle_single_event(event, backends)
        self.assertEquals(1, len(captured))
        del captured[:]

        event = SystemEvent(kind=SystemEvent.KEG_VOLUME_LOW)
        notification.handle_single_event(event, backends)
        self.assertEquals(0, len(captured))
        prefs.keg_volume_low = True
        prefs.save()
        notification.handle_single_event(event, backends)
        self.assertEquals(1, len(captured))
        del captured[:]




### Constants

# Internal version number, bumped every time "kb_upgrade" is needed.
EPOCH = 103

try:
    import local_settings
    IS_SETUP = True
except ImportError:
    IS_SETUP = False

### Celery setup
# This will make sure the app is always imported when
# Django starts so that shared_task will use this app.
# HACK: Avoid bootstrapping Celery until setup has been run,
# since it depends on settings -> local_settings.
if IS_SETUP:
    from .celery import app as celery_app


from django import forms
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Div, Submit, HTML, Button, Row, Field, Hidden, Div
from crispy_forms.bootstrap import AppendedText, PrependedText, FormActions

from kegbot.util import units
from pykeg.core import backend
from pykeg.core import models

ALL_TAPS = models.KegTap.objects.all()
ALL_SIZES = models.KegSize.objects.all()
ALL_KEGS = models.Keg.objects.all()

class GeneralSettingsForm(forms.Form):
    name = forms.CharField(help_text='Name of this Kegbot system')

class ChangeKegForm(forms.Form):
    keg_size = forms.ModelChoiceField(queryset=ALL_SIZES,
        required=True,
        empty_label=None)

    beer_name = forms.CharField(required=True, label='Beer Name')
    beer_id = forms.CharField(widget=forms.HiddenInput(), required=False)
    brewer_name = forms.CharField(required=True, label='Brewer')
    brewer_id = forms.CharField(widget=forms.HiddenInput(), required=False)
    style_name = forms.CharField(required=True, label='Style',
      help_text='Example: Pale Ale, Stout, etc.')
    style_id = forms.CharField(widget=forms.HiddenInput(), required=False)

    helper = FormHelper()
    helper.form_class = 'form-horizontal beer-select'
    helper.layout = Layout(
        Field('beer_name', css_class='input-xlarge'),
        Field('beer_id', type='hidden'),
        Field('brewer_name', css_class='input-xlarge'),
        Field('brewer_id', type='hidden'),
        Field('style_name', css_class='input-xlarge'),
        Field('style_id', type='hidden'),
        Field('keg_size', css_class='input-xlarge'),
        FormActions(
            Submit('submit_change_keg_form', 'Activate Keg', css_class='btn-primary'),
        )
    )

    def save(self, tap):
        if not self.is_valid():
            raise ValueError('Form is not valid.')
        b = backend.KegbotBackend()

        if tap.is_active():
            b.end_keg(tap)

        cd = self.cleaned_data
        keg = b.start_keg(tap, beer_name=cd['beer_name'], brewer_name=cd['brewer_name'],
            style_name=cd['style_name'], keg_size=cd['keg_size'])

        if cd.get('description'):
            keg.description = cd['description']
            keg.save()


class EndKegForm(forms.Form):
    keg = forms.ModelChoiceField(queryset=ALL_KEGS, widget=forms.HiddenInput)

    helper = FormHelper()
    helper.form_class = 'form-horizontal beer-select'
    helper.layout = Layout(
        Field('keg', type='hidden'),
        FormActions(
          Submit('submit_end_keg_form', 'End Keg', css_class='btn-danger'),
        )
    )


class TapForm(forms.ModelForm):
    class Meta:
        model = models.KegTap
        fields = ('name', 'meter_name', 'relay_name', 'description',
            'temperature_sensor', 'ml_per_tick')

    def __init__(self, *args, **kwargs):
        site = kwargs.pop('site', None)
        super(TapForm, self).__init__(*args, **kwargs)
        self.fields['temperature_sensor'].queryset = models.ThermoSensor.objects.all()
        self.fields['temperature_sensor'].empty_label = 'No sensor.'

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('name', css_class='input-xlarge'),
        Field('meter_name', css_class='input-xlarge'),
        Field('relay_name', css_class='input-xlarge'),
        Field('temperature_sensor', css_class='input-xlarge'),
        Field('ml_per_tick', css_class='input-xlarge'),
        Field('description', css_class='input-xlarge'),
        FormActions(
            Submit('submit_tap_form', 'Save Settings', css_class='btn-success'),
        )
    )


class DeleteTapForm(forms.Form):
    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        FormActions(
            Submit('submit_delete_tap_form', 'Delete Tap', css_class='btn-danger'),
        )
    )


class SiteSettingsForm(forms.ModelForm):
    class Meta:
        model = models.SiteSettings
        fields = (
            'title',
            'privacy',
            'volume_display_units',
            'temperature_display_units',
            'timezone',
            'hostname',
            'use_ssl',
            'session_timeout_minutes',
            'google_analytics_id',
            'guest_name',
            'guest_image',
            'default_user',
            'registration_allowed',
            'registration_confirmation',
            'allowed_hosts',
        )

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('title', css_class='input-xlarge'),
        Field('privacy', css_class='input-xlarge'),
        Field('volume_display_units', css_class='input-xlarge'),
        Field('temperature_display_units', css_class='input-xlarge'),
        Field('timezone'),
        Field('hostname'),
        Field('use_ssl'),
        Field('session_timeout_minutes'),
        Field('google_analytics_id'),
        Field('guest_name'),
        Field('guest_image'),
        Field('default_user'),
        Field('registration_allowed', css_class='input-xlarge'),
        Field('registration_confirmation', css_class='input-xlarge'),
        Field('allowed_hosts'),
        FormActions(
            Submit('submit', 'Save Settings', css_class='btn-primary'),
        )
    )

class BeerTypeForm(forms.ModelForm):
    class Meta:
        model = models.BeerType
        fields = ('name', 'style', 'brewer', 'edition', 'abv',
            'calories_oz', 'carbs_oz', 'original_gravity', 'specific_gravity',
            'untappd_beer_id')

    new_image = forms.ImageField(required=False,
        help_text='Set/replace image for this beer type.')

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('name', css_class='input-xlarge'),
        Field('style', css_class='input-xlarge'),
        Field('brewer'),
        Field('edition'),
        Field('abv'),
        Field('calories_oz'),
        Field('carbs_oz'),
        Field('original_gravity'),
        Field('specific_gravity'),
        Field('untappd_beer_id'),
        Field('new_image'),
        FormActions(
            Submit('submit', 'Save', css_class='btn-primary'),
        )
    )

class BrewerForm(forms.ModelForm):
    class Meta:
        model = models.Brewer

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('name', css_class='input-xlarge'),
        Field('country', css_class='input-xlarge'),
        Field('origin_state', css_class='input-xlarge'),
        Field('origin_city', css_class='input-xlarge'),
        Field('production'),
        Field('url'),
        Field('description'),
        Field('image'),
        FormActions(
            Submit('submit', 'Save', css_class='btn-primary'),
        )
    )

class BeerStyleForm(forms.ModelForm):
    class Meta:
        model = models.BeerStyle

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('name', css_class='input-xlarge'),
        FormActions(
            Submit('submit', 'Save', css_class='btn-primary'),
        )
    )

class FindUserForm(forms.Form):
    username = forms.CharField()

class UserForm(forms.ModelForm):
    class Meta:
        model = models.User
        fields = (
            'username',
            'email',
            'password',
            'is_staff',
        )

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('username', css_class='input-xlarge'),
        Field('email', css_class='input-xlarge'),
        'password',
        'is_staff',
        FormActions(
            Submit('submit', 'Save', css_class='btn-primary'),
        )
    )

class TokenForm(forms.ModelForm):
    class Meta:
        model = models.AuthenticationToken
        fields = (
            'nice_name',
            'enabled',
        )
    username = forms.CharField(required=False)

    helper = FormHelper()
    helper.form_class = 'form-horizontal user-select'
    helper.layout = Layout(
        Field('username', css_class='input-xlarge'),
        Field('nice_name', css_class='input-xlarge'),
        'enabled',
        FormActions(
            Submit('submit', 'Save', css_class='btn-primary'),
        )
    )

    def clean_username(self):
        username = self.cleaned_data['username']
        if username == '':
            self.cleaned_data['user'] = None
            return
        try:
            self.cleaned_data['user'] = models.User.objects.get(username=username)
        except models.User.DoesNotExist:
            raise forms.ValidationError('Invalid username; use a complete user name or leave blank.')
        return username


class AddTokenForm(forms.ModelForm):
    class Meta:
        model = models.AuthenticationToken
        fields = (
            'auth_device',
            'token_value',
            'enabled',
        )
    CHOICES = (
      ('core.rfid', 'RFID'),
      ('core.onewire', 'OneWire/iButton'),
      ('nfc', 'NFC'),
    )
    auth_device = forms.ChoiceField(choices=CHOICES)
    username = forms.CharField(required=False)

    helper = FormHelper()
    helper.form_class = 'form-horizontal user-select'
    helper.layout = Layout(
        Field('auth_device', css_class='input-xlarge'),
        Field('token_value', css_class='input-xlarge'),
        Field('username', css_class='input-xlarge'),
        'enabled',
        FormActions(
            Submit('submit', 'Save', css_class='btn-primary'),
        )
    )

    def clean_username(self):
        username = self.cleaned_data['username']
        if username == '':
            self.cleaned_data['user'] = None
            return
        try:
            self.cleaned_data['user'] = models.User.objects.get(username=username)
        except models.User.DoesNotExist:
            raise forms.ValidationError('Invalid username; use a complete user name or leave blank.')
        return username


class CancelDrinkForm(forms.Form):
    pass

class ReassignDrinkForm(forms.Form):
    username = forms.CharField(required=False)

class ChangeDrinkVolumeForm(forms.Form):
    UNIT_CHOICES = (
      ('mL', 'mL'),
      ('oz', 'oz')
    )
    units = forms.ChoiceField(required=True, choices=UNIT_CHOICES)
    volume = forms.FloatField(required=True, min_value=0)

    def clean_volume(self):
        volume = self.cleaned_data['volume']
        if self.cleaned_data['units'] == 'oz':
            self.cleaned_data['volume_ml'] = float(units.Quantity(volume, units.UNITS.Ounce).InMilliliters())
        else:
            self.cleaned_data['volume_ml'] = volume
        return volume

class RecordDrinkForm(forms.Form):
    units = forms.ChoiceField(required=True, choices=ChangeDrinkVolumeForm.UNIT_CHOICES)
    volume = forms.FloatField(required=True, min_value=0)
    username = forms.CharField(required=False)

    def clean_username(self):
        username = self.cleaned_data['username']
        if username == '':
            self.cleaned_data['user'] = None
            return
        try:
            self.cleaned_data['user'] = models.User.objects.get(username=username)
        except models.User.DoesNotExist:
            raise forms.ValidationError('Invalid username; use a complete user name or leave blank.')
        return username

    def clean_volume(self):
        volume = self.cleaned_data['volume']
        if self.cleaned_data['units'] == 'oz':
            self.cleaned_data['volume_ml'] = float(units.Quantity(volume, units.UNITS.Ounce).InMilliliters())
        else:
            self.cleaned_data['volume_ml'] = volume
        return volume



# Copyright 2011 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import datetime

from django.utils import timezone
from kegbot.util import units

from pykeg.core import models

def to_pints(volume):
    return float(units.Quantity(volume).InPints())

class ChartError(Exception):
    """Base chart exception."""

def TemperatureSensorChart(sensor):
    """ Shows a simple line plot of a specific temperature sensor.

    Syntax:
      {% chart sensor <sensorname> width height %}
    Args:
      sensorname - the nice_name of a ThermoSensor
    """
    if not isinstance(sensor, models.ThermoSensor):
        raise ChartError('Bad sensor given')

    hours = 6
    now = timezone.now()
    start = now - (datetime.timedelta(hours=hours))
    start = start - (datetime.timedelta(seconds=start.second))

    points = sensor.thermolog_set.filter(time__gte=start).order_by('time')

    curr = start
    temps = []
    have_temps = False
    for point in points:
        while curr <= point.time:
            curr += datetime.timedelta(minutes=1)
            if curr < point.time:
                temps.append(None)
            else:
                temps.append(point.temp)
                have_temps = True

    if not have_temps:
        raise ChartError('Not enough data')

    res = {
      'series': [
        {
          'data': temps,
          'marker': {
            'enabled': False,
          },
        },
      ],
      'tooltip': {
        'enabled': False,
      },
      'xAxis': {
        'categories': ['Temperature'],
        'labels': {
          'enabled': False,
        },
        'tickInterval': 0,
      },
      'yAxis': {
        'labels': {
          'enabled': False,
        },
        'tickInterval': 1,
      },
    }
    return res

def VolumeByWeekday(stats):
    """ Shows a histogram of volume by day of the week.

    Syntax:
      {% chart volume_by_day <stats> width height %}
    Args:
      stats - a stats object containing volume_by_day_of_week
    """
    volmap = [0] * 7
    vols = stats.get('volume_by_day_of_week', {})
    if not volmap:
        raise ChartError('Daily volumes unavailable')

    for weekday, vol in vols.iteritems():
        volmap[int(weekday)] += to_pints(vol)
    return _DayOfWeekChart(volmap)

def UserSessionsByWeekday(user):
    """Shows a user's total session by volume by day of week."""
    if not isinstance(user, models.User):
        raise ChartError('Bad value for user')
    chunks = user.user_session_chunks.all()
    weekdays = [0] * 7
    for chunk in chunks:
        # Convert from Sunday = 6 to Sunday = 0
        weekday = (chunk.start_time.weekday() + 1) % 7
        weekdays[weekday] += to_pints(chunk.volume_ml)
    return _DayOfWeekChart(weekdays)

def SessionVolumes(sessions):
    buckets = [0]*6
    labels = [
      '<1',
      '1.0-1.9',
      '2.0-2.9',
      '3.0-3.9',
      '4.0-4.9',
      '5+'
    ]
    for sess in sessions:
        pints = round(to_pints(sess.volume_ml), 1)
        intval = int(pints)
        if intval >= len(buckets):
            buckets[-1] += 1
        else:
            buckets[intval] += 1

    res = {
      'xAxis': {
        'categories': labels,
      },
      'series': [
        {'data': buckets},
      ],
      'yAxis': {
        'min': 0,
      },
      'chart': {
        'defaultSeriesType': 'column',
      }
    }
    return res

def UsersByVolume(stats):
    vols = stats.get('volume_by_drinker')
    if not vols:
        raise ChartError('no data')

    data = []
    for username, volume in vols.iteritems():
        pints = to_pints(volume)
        label = '<b>%s</b> (%.1f)' % (username, pints)
        data.append((label, pints))

    other_vol = 0
    for username, pints in data[10:]:
        other_vol += pints

    def _sort_vol_desc(a, b):
        return cmp(b[1], a[1])

    data.sort(_sort_vol_desc)
    data = data[:10]
    data.reverse()

    if other_vol:
        label = '<b>%s</b> (%.1f)' % ('all others', other_vol)
        data.append((label, other_vol))

    res = {
      'series': [
        {
          'type': 'pie',
          'name': 'Drinkers by Volume',
          'data': data,
        }
      ],
      'yAxis': {
        'min': 0,
      },
      'chart': {
        'defaultSeriesType': 'column',
      },
      'tooltip': {
        'enabled': False,
      },
    }
    return res

def _DayOfWeekChart(vals):
    labels = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat']

    # convert from 0=Monday to 0=Sunday
    #vals.insert(0, vals.pop(-1))

    res = {
      'xAxis': {
        'categories': labels,
      },
      'yAxis': {
        'min': 0,
      },
      'series': [
        {'data': vals},
      ],
      'tooltip': {
        'enabled': False,
      },
      'chart': {
        'defaultSeriesType': 'column',
      }
    }
    return res



# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django import forms
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Submit, Field
from crispy_forms.bootstrap import FormActions

from pykeg.core import models

class CreateOrImportForm(forms.Form):
    CHOICES = (
        ('create', 'Create New System'),
        ('import', 'Import From Backup'),
    )
    mode = forms.ChoiceField(widget=forms.RadioSelect, choices=CHOICES)
    import_file  = forms.FileField(required=False)

    helper = FormHelper()
    helper.form_class = 'setup-form span8 offset2'
    helper.layout = Layout(
        Field('mode', css_class='input-xlarge'),
        Field('import_file', css_class='input-xlarge'),
        FormActions(
            Submit('save_changes', 'Continue', css_class="btn-primary"),
        )
    )


class MiniSiteSettingsForm(forms.ModelForm):
    class Meta:
        model = models.SiteSettings
        fields = (
            'title',
            'privacy',
            'timezone',
            'hostname',
            'volume_display_units',
            'temperature_display_units',
            'registration_allowed',
            'registration_confirmation',
        )
        widgets = {
            'privacy': forms.RadioSelect(),
            'volume_display_units': forms.RadioSelect(),
            'temperature_display_units': forms.RadioSelect(),
        }

    helper = FormHelper()
    helper.form_class = 'setup-form span8 offset2'
    helper.layout = Layout(
        Field('title', css_class='span12'),
        Field('privacy', css_class='span12'),
        Field('timezone', css_class='span12'),
        Field('hostname', css_class='span12'),
        Field('volume_display_units', css_class='span12'),
        Field('temperature_display_units', css_class='span12'),
        Field('registration_allowed'),
        Field('registration_confirmation'),
        FormActions(
            Submit('save_changes', 'Continue', css_class="btn-primary"),
        )
    )

class AdminUserForm(forms.Form):
    username = forms.RegexField(
        label='Username',
        max_length=30,
        regex=r'^[\w-]+$',
        help_text='Your username: 30 characters or fewer. Alphanumeric characters only (letters, digits, hyphens and underscores).',
        error_message='Must contain only letters, numbers, hyphens and underscores.')

    password = forms.CharField(widget=forms.PasswordInput())
    confirm_password = forms.CharField(widget=forms.PasswordInput())
    email = forms.EmailField(
        help_text='In case you lose your password.')

    helper = FormHelper()
    helper.form_class = 'setup-form span8 offset2'
    helper.layout = Layout(
        Field('username', css_class='span12'),
        Field('email', css_class='span12'),
        Field('password', css_class='span12'),
        Field('confirm_password', css_class='span12'),
        FormActions(
            Submit('save_changes', 'Continue', css_class="btn-primary"),
        )
    )

    def clean_confirm_password(self):
        orig = self.cleaned_data.get('password')
        confirm = self.cleaned_data.get('confirm_password')
        if orig and confirm and orig != confirm:
            raise forms.ValidationError('Passwords must match!')
        return confirm

    def clean_username(self):
        username = self.cleaned_data.get('username')
        if username:
            try:
                models.User.objects.get(username=username)
                raise forms.ValidationError('Sorry, this username is taken already?!')
            except models.User.DoesNotExist:
                pass  # expected
        return username

    def save(self):
        u = models.User()
        u.username = self.cleaned_data.get('username')
        u.email = self.cleaned_data.get('email')
        u.is_staff = True
        u.is_superuser = True
        u.save()
        u.set_password(self.cleaned_data.get('password'))
        u.save()
        return u



# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for pykeg.core.models"""

import datetime

from django.conf import settings
from django.utils import timezone
from django.test import TestCase

from . import backend
from . import kb_common
from . import models
from .testutils import make_datetime

from kegbot.util import units

class CoreModelsTestCase(TestCase):
    def setUp(self):
        models.KegbotSite.get()  # create the site
        self.backend = backend.KegbotBackend()
        self.brewer = models.Brewer.objects.create(
            name='Moonshine Beers',
            country='USA',
            origin_state='Anystate',
            origin_city='Bathtub',
            production='retail',
            url='http://example.com/',
            description='Pretty bad beers.',
        )

        self.beer_style = models.BeerStyle.objects.create(
            name='Porter',
        )

        self.beer_type = models.BeerType.objects.create(
            name='Moonshine Porter',
            brewer=self.brewer,
            style=self.beer_style,
            calories_oz=3.0,
            carbs_oz=10.0,
            abv=0.05,
        )

        self.keg_size = models.KegSize.objects.create(
            name='Tiny Keg',
            volume_ml=2000,
        )

        self.keg = models.Keg.objects.create(
            type=self.beer_type,
            size=self.keg_size,
            start_time=make_datetime(2000, 4, 1),
            end_time=make_datetime(2000, 5, 1),
            description='Our first keg!',
            full_volume_ml=2000,
        )

        self.tap = models.KegTap.objects.create(
            name='Test Tap',
            meter_name='test',
            ml_per_tick=(1000.0/2200.0),
            current_keg=self.keg,
        )

        self.user = models.User.objects.create(
            username='kb_tester',
        )

        self.user2 = models.User.objects.create(
            username='kb_tester2',
        )

    def tearDown(self):
        self.user.delete()
        self.user2.delete()
        self.keg.delete()
        self.beer_type.delete()
        self.beer_style.delete()
        self.brewer.delete()

    def testKegStuff(self):
        """Test basic keg relations that should always work."""
        self.assertEqual(self.keg.size.volume_ml,
            units.Quantity(2.0, units.UNITS.Liter).InMilliliters())
        self.assertEqual(self.keg.type.brewer.name, "Moonshine Beers")

        self.assertEqual(0.0, self.keg.served_volume())
        self.assertEqual(2000, self.keg.remaining_volume())

    def testDrinkAccounting(self):
        vol = units.Quantity(1200)

        d = self.backend.record_drink(tap_name=self.tap.meter_name,
            ticks=1200,
            username=self.user.username,
        )

        self.assertEqual(d.keg.served_volume(), d.volume_ml)

    def testDrinkSessions(self):
        """ Checks for the DrinkingSession records. """
        u1 = self.user
        u2 = self.user2
        vol = units.Quantity(1200)

        drinks = {}
        base_time = make_datetime(2009,1,1,1,0,0)

        ticks = volume = vol.InKbMeterTicks()

        td_10m = datetime.timedelta(minutes=10)
        td_400m = datetime.timedelta(minutes=400)
        td_390m = td_400m - td_10m

        self.assertEqual(models.Drink.objects.all().count(), 0)
        self.assertEqual(models.DrinkingSession.objects.all().count(), 0)

        # u=1 t=0
        self.backend.record_drink(tap_name=self.tap.meter_name,
            ticks=1200,
            username=u1.username,
            pour_time=base_time,
        )
        # u=2 t=0
        self.backend.record_drink(tap_name=self.tap.meter_name,
            ticks=1200,
            username=u2.username,
            pour_time=base_time,
        )

        # u=1 t=10
        self.backend.record_drink(tap_name=self.tap.meter_name,
            ticks=1200,
            username=u1.username,
            pour_time=base_time+td_10m,
        )

        # u=1 t=400
        self.backend.record_drink(tap_name=self.tap.meter_name,
            ticks=1200,
            username=u1.username,
            pour_time=base_time+td_400m,
        )

        # u=2 t=490
        self.backend.record_drink(tap_name=self.tap.meter_name,
            ticks=1200,
            username=u2.username,
            pour_time=base_time+td_390m,
        )

        # u=2 t=400
        self.backend.record_drink(tap_name=self.tap.meter_name,
            ticks=1200,
            username=u2.username,
            pour_time=base_time+td_400m,
        )

        drinks_u1 = u1.drinks.all().order_by('time')
        drinks_u2 = u2.drinks.all().order_by('time')

        u1_chunks = u1.session_chunks.all().order_by('start_time')
        self.assertEqual(len(u1_chunks), 2)

        u2_chunks = u2.session_chunks.all().order_by('start_time')
        self.assertEqual(len(u2_chunks), 2)

        s1, s2 = models.DrinkingSession.objects.all().order_by('start_time')[:2]

        SESSION_DELTA = datetime.timedelta(minutes=kb_common.DRINK_SESSION_TIME_MINUTES)

        # session 1: should be 10 minutes long as created above
        self.assertEqual(s1.start_time, drinks_u1[0].time)
        self.assertEqual(s1.end_time, drinks_u1[0].time + td_10m + SESSION_DELTA)
        self.assertEqual(s1.drinks.all().filter(user=u1).count(), 2)
        self.assertEqual(s1.drinks.all().filter(user=u2).count(), 1)

        # session 2: at time 200, 1 drink
        self.assertEqual(s2.start_time, base_time + td_390m)
        self.assertEqual(s2.end_time, base_time + td_400m + SESSION_DELTA)
        self.assertEqual(s2.drinks.all().filter(user=u1).count(), 1)
        self.assertEqual(s2.drinks.all().filter(user=u2).count(), 2)

        # user2 session2: drinks are added out of order to create this, ensure times
        # match
        u2_c2 = u2_chunks[1]
        self.assertEqual(u2_c2.start_time, base_time + td_390m)
        self.assertEqual(u2_c2.end_time, base_time + td_400m + SESSION_DELTA)

        # Now check DrinkingSessions were created correctly; there should be
        # two groups capturing all 4 sessions.
        all_groups = models.DrinkingSession.objects.all().order_by('start_time')
        self.assertEqual(len(all_groups), 2)

        self.assertEqual(all_groups[0].start_time, base_time)
        self.assertEqual(all_groups[0].end_time, base_time + td_10m + SESSION_DELTA)
        self.assertEqual(all_groups[0].user_chunks.all().count(), 2)

        self.assertEqual(all_groups[1].start_time, base_time + td_390m)
        self.assertEqual(all_groups[1].end_time, base_time + td_400m + SESSION_DELTA)
        self.assertEqual(all_groups[1].user_chunks.all().count(), 2)



# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""System constants and defaults."""

# Where to look for local_settings.py
LOCAL_SETTINGS_SEARCH_DIRS = (
    '~/.kegbot/',
    '/etc/kegbot/',
    '/usr/local/etc/kegbot',
)

# SF800 default (5400 ticks/Liter == 5.4 ticks/mL == 1/5.4 mL/tick)
ML_PER_TICK = (1/5.4)

# Minimum session size, in milliliters, to display
MIN_SESSION_VOLUME_DISPLAY_ML = 177  # 6 ounces

# Maximum time between consecutive drinks to be considered in the same 'session'
# (see UserDrinkingSession table)
DRINK_SESSION_TIME_MINUTES = 3 * 60

# Minimum and maximum thermo sensor readings (degrees C).
THERMO_SENSOR_RANGE = (-20.0, 80.0)

# Maximum number of readings to keep.
THERMO_SENSOR_HISTORY_MINUTES = 60 * 24

# Device names
AUTH_MODULE_CORE_ONEWIRE = 'core.onewire'
AUTH_MODULE_CORE_RFID = 'core.rfid'
AUTH_MODULE_CONTRIB_PHIDGET_RFID = AUTH_MODULE_CORE_RFID

# Auth modules whose token values should be interpreted as lower-case hex.
AUTH_MODULE_NAMES_HEX_VALUES = (AUTH_MODULE_CORE_ONEWIRE, AUTH_MODULE_CORE_RFID)

# Low volume threshold: 15% full
KEG_VOLUME_LOW_PERCENT = 0.15




# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.db import connection
from django.core.management.base import CommandError
from django.core.management.base import NoArgsCommand

from pykeg.core import models
from pykeg.core.management.commands.common import progbar

class Command(NoArgsCommand):
    help = u'Regenerate all cached stats.'
    args = '<none>'

    def handle(self, **options):
        drinks = models.Drink.objects.all()

        pos = 0
        count = drinks.count()
        for d in drinks:
            pos += 1
            progbar('clear drink sessions', pos, count)
            d.session = None
            d.save()
        print ''

        pics = models.PourPicture.objects.all()
        count = pics.count()
        pos = 0
        for p in pics:
            pos += 1
            progbar('clear image sessions', pos, count)
            p.session = None
            p.save()
        print ''

        print 'deleting old sessions..',
        try:
            with connection.constraint_checks_disabled():
                cursor = connection.cursor()
                for table in ('core_drinkingsession', 'core_kegsessionchunk',
                  'core_usersessionchunk', 'core_sessionchunk'):
                    cursor.execute('TRUNCATE TABLE `%s`' % table)
                print 'truncate successful'
        except Exception, e:
            models.DrinkingSession.objects.all().delete()
            print 'orm delete successful'

        pos = 0
        count = drinks.count()
        for d in drinks.order_by('time'):
            pos += 1
            progbar('calc new sessions', pos, count)
            sess = models.DrinkingSession.AssignSessionForDrink(d)
        print ''

        pics = models.PourPicture.objects.all()
        count = pics.count()
        pos = 0
        for p in pics:
            pos += 1
            progbar('set image sessions', pos, count)
            if p.drink:
                p.session = p.drink.session
                p.save()
        print ''

        print 'done!'



# Copyright 2012 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

'''Tests for middleware.py.'''

import unittest

from . import middleware

class MiddlewareTestCase(unittest.TestCase):
    def test_validate_host(self):
        allowed_hosts = []
        self.assertTrue(middleware.HttpHostMiddleware.validate_host('foo', allowed_hosts))

        allowed_hosts = ['blort', '*']
        self.assertTrue(middleware.HttpHostMiddleware.validate_host('foo', allowed_hosts))

        allowed_hosts = ['blort']
        self.assertFalse(middleware.HttpHostMiddleware.validate_host('foo', allowed_hosts))

        # Patterns are not supported.. yet.
        allowed_hosts = ['*.bar.com']
        self.assertFalse(middleware.HttpHostMiddleware.validate_host('foo.bar.com', allowed_hosts))



#!/usr/bin/env python
#
# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import cStringIO
import datetime

from django.contrib import messages
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.core.urlresolvers import reverse
from django.db.models import Q
from django.http import Http404
from django.http import HttpResponse
from django.shortcuts import get_object_or_404
from django.shortcuts import redirect
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.forms import widgets
from django.views.decorators.http import require_POST
from django.utils import timezone
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger
from django.views.decorators.http import require_http_methods

from kegbot.util import kbjson
from kegbot.util import units

from pykeg.core import backend
from pykeg.core import logger
from pykeg.core import models

from pykeg.web.kegadmin import forms

@staff_member_required
def dashboard(request):
    context = RequestContext(request)
    recent_time = timezone.now() - datetime.timedelta(days=30)
    active_users = models.User.objects.filter(is_active=True)
    new_users = models.User.objects.filter(date_joined__gte=recent_time)
    context['num_users'] = len(active_users)
    context['num_new_users'] = len(new_users)
    return render_to_response('kegadmin/dashboard.html', context_instance=context)

@staff_member_required
def general_settings(request):
    context = RequestContext(request)
    settings = models.SiteSettings.get()
    form = forms.SiteSettingsForm(instance=settings)
    if request.method == 'POST':
        form = forms.SiteSettingsForm(request.POST, instance=settings)
        if form.is_valid():
            form.save()
            messages.success(request, 'Site settings were successfully updated.')
    context['settings_form'] = form
    return render_to_response('kegadmin/index.html', context_instance=context)

@staff_member_required
def tap_list(request):
    context = RequestContext(request)
    context['taps'] = models.KegTap.objects.all()
    return render_to_response('kegadmin/tap_list.html', context_instance=context)

@staff_member_required
def add_tap(request):
    context = RequestContext(request)
    form = forms.TapForm(site=request.kbsite)
    if request.method == 'POST':
        form = forms.TapForm(request.POST, site=request.kbsite)
        if form.is_valid():
            new_tap = form.save(commit=False)
            new_tap.site = request.kbsite
            new_tap.save()
            messages.success(request, 'Tap created.')
            return redirect('kegadmin-taps')
    context['form'] = form
    return render_to_response('kegadmin/add_tap.html', context_instance=context)

@staff_member_required
def tap_detail(request, tap_id):
    tap = get_object_or_404(models.KegTap, id=tap_id)

    record_drink_form = forms.RecordDrinkForm()
    activate_keg_form = forms.ChangeKegForm()
    tap_settings_form = forms.TapForm(instance=tap, site=request.kbsite)

    if request.method == 'POST':
        if 'submit_change_keg_form' in request.POST:
            activate_keg_form = forms.ChangeKegForm(request.POST)
            if activate_keg_form.is_valid():
                activate_keg_form.save(tap)
                messages.success(request, 'The new keg was activated. Bottoms up!')
                return redirect('kegadmin-taps')

        elif 'submit_tap_form' in request.POST:
            tap_settings_form = forms.TapForm(request.POST, instance=tap, site=request.kbsite)
            if tap_settings_form.is_valid():
                tap_settings_form.save()
                messages.success(request, 'Tap settings saved.')
                tap_settings_form = forms.TapForm(instance=tap, site=request.kbsite)

        elif 'submit_delete_tap_form' in request.POST:
            delete_form = forms.DeleteTapForm(request.POST)
            if delete_form.is_valid():
                if tap.current_keg:
                    request.backend.end_keg(tap)
                tap.delete()
                messages.success(request, 'Tap deleted.')
                return redirect('kegadmin-taps')

        elif 'submit_end_keg_form' in request.POST:
            end_keg_form = forms.EndKegForm(request.POST)
            if end_keg_form.is_valid():
                old_keg = request.backend.end_keg(tap)
                messages.success(request, 'Keg %s was ended.' % old_keg.id)

        elif 'submit_record_drink' in request.POST:
            record_drink_form = forms.RecordDrinkForm(request.POST)
            if record_drink_form.is_valid():
                user = record_drink_form.cleaned_data.get('user')
                volume_ml = record_drink_form.cleaned_data.get('volume_ml')
                d = request.backend.record_drink(tap.meter_name, ticks=0, username=user,
                    volume_ml=volume_ml)
                messages.success(request, 'Drink %s recorded.' % d.id)
            else:
                messages.error(request, 'Please enter a valid volume and user.')

        else:
            messages.warning(request, 'No form data was found. Bug?')

    end_keg_form = forms.EndKegForm(initial={'keg':tap.current_keg})

    context = RequestContext(request)
    context['tap'] = tap
    context['current_keg'] = tap.current_keg
    context['activate_keg_form'] = activate_keg_form
    context['record_drink_form'] = record_drink_form
    context['end_keg_form'] = end_keg_form
    context['tap_settings_form'] = tap_settings_form
    context['delete_tap_form'] = forms.DeleteTapForm()
    return render_to_response('kegadmin/tap_detail.html', context_instance=context)

@staff_member_required
def user_list(request):
    context = RequestContext(request)

    if request.method == 'POST':
        form = forms.FindUserForm(request.POST)
        if form.is_valid():
            username = form.cleaned_data.get('username')
            try:
                user = models.User.objects.get(username=username)
                return redirect('kegadmin-edit-user', user.id)
            except models.User.DoesNotExist:
                messages.error(request, 'User "%s" does not exist.' % username)

    users = models.User.objects.all().order_by('-id')
    paginator = Paginator(users, 25)

    page = request.GET.get('page')
    try:
        users = paginator.page(page)
    except PageNotAnInteger:
        users = paginator.page(1)
    except EmptyPage:
        users = paginator.page(paginator.num_pages)

    context['users'] = users
    return render_to_response('kegadmin/user_list.html', context_instance=context)

@staff_member_required
def add_user(request):
    context = RequestContext(request)
    form = forms.UserForm()
    if request.method == 'POST':
        form = forms.UserForm(request.POST)
        if form.is_valid():
            instance = form.save(commit=False)
            instance.set_password(form.cleaned_data.get('password'))
            instance.save()
            messages.success(request, 'User "%s" created.' % instance.username)
            return redirect('kegadmin-users')
    context['form'] = form
    return render_to_response('kegadmin/add_user.html', context_instance=context)

@staff_member_required
def user_detail(request, user_id):
    user = get_object_or_404(models.User, id=user_id)

    if request.method == 'POST':
        if 'submit_enable' in request.POST:
            if user.is_active:
                messages.error(request, 'User is already enabled.')
            else:
                user.is_active = True
                user.save()
                messages.success(request, 'User %s was enabled.' % user.username)

        elif 'submit_disable' in request.POST:
            if not user.is_active:
                messages.error(request, 'User is already disabled.')
            else:
                user.is_active = False
                user.save()
                messages.success(request, 'User %s was disabled.' % user.username)

    context = RequestContext(request)
    context['user'] = user
    context['profile'] = user.get_profile()
    context['stats'] = context['profile'].GetStats()

    context['tokens'] = user.tokens.all().order_by('created_time')

    return render_to_response('kegadmin/user_detail.html', context_instance=context)

@staff_member_required
def drink_list(request):
    context = RequestContext(request)
    drinks = models.Drink.objects.all().order_by('-time')
    paginator = Paginator(drinks, 25)

    page = request.GET.get('page')
    try:
        drinks = paginator.page(page)
    except PageNotAnInteger:
        drinks = paginator.page(1)
    except EmptyPage:
        drinks = paginator.page(paginator.num_pages)

    context['drinks'] = drinks
    return render_to_response('kegadmin/drink_list.html', context_instance=context)


@staff_member_required
@require_http_methods(["POST"])
def drink_edit(request, drink_id):
    drink = get_object_or_404(models.Drink, id=drink_id)

    if 'submit_cancel' in request.POST:
        form = forms.CancelDrinkForm(request.POST)
        old_keg = drink.keg
        if form.is_valid():
            request.backend.cancel_drink(drink)
            messages.success(request, 'Drink %s was cancelled.' % drink_id)
            return redirect(old_keg.get_absolute_url())
        else:
            messages.error(request, 'Invalid request')
            return redirect(drink.get_absolute_url())

    elif 'submit_reassign' in request.POST:
        form = forms.ReassignDrinkForm(request.POST)
        if form.is_valid():
            username = form.cleaned_data.get('username', None)
            try:
                request.backend.assign_drink(drink, username)
                messages.success(request, 'Drink %s was reassigned.' % drink_id)
            except models.User.DoesNotExist:
                messages.error(request, 'No such user')
        else:
            messages.error(request, 'Invalid request')
        return redirect(drink.get_absolute_url())

    elif 'submit_edit_volume' in request.POST:
        form = forms.ChangeDrinkVolumeForm(request.POST)
        if form.is_valid():
            volume_ml = form.cleaned_data.get('volume_ml')
            if volume_ml == drink.volume_ml:
                messages.warning(request, 'Drink volume unchanged.')
            else:
                request.backend.set_drink_volume(drink, volume_ml)
                messages.success(request, 'Drink %s was updated.' % drink_id)
        else:
            messages.error(request, 'Please provide a valid volume.')
        return redirect(drink.get_absolute_url())

    message.error(request, 'Unknown action.')
    return redirect(drink.get_absolute_url())


@staff_member_required
def token_list(request):
    context = RequestContext(request)
    tokens = models.AuthenticationToken.objects.all().order_by('-created_time')
    paginator = Paginator(tokens, 25)

    page = request.GET.get('page')
    try:
        tokens = paginator.page(page)
    except PageNotAnInteger:
        tokens = paginator.page(1)
    except EmptyPage:
        tokens = paginator.page(paginator.num_pages)

    context['tokens'] = tokens
    return render_to_response('kegadmin/token_list.html', context_instance=context)


@staff_member_required
def token_detail(request, token_id):
    token = get_object_or_404(models.AuthenticationToken, id=token_id)

    username = ''
    if token.user:
        username = token.user.username

    form = forms.TokenForm(instance=token, initial={'username': username})
    if request.method == 'POST':
        form = forms.TokenForm(request.POST, instance=token)
        if form.is_valid():
            instance = form.save(commit=False)
            instance.user = form.cleaned_data['user']
            instance.save()
            messages.success(request, 'Token updated.')

    context = RequestContext(request)
    context['token'] = token
    context['form'] = form
    return render_to_response('kegadmin/token_detail.html', context_instance=context)

@staff_member_required
def add_token(request):
    context = RequestContext(request)
    form = forms.AddTokenForm()
    if request.method == 'POST':
        form = forms.AddTokenForm(request.POST)
        if form.is_valid():
            instance = form.save(commit=False)
            instance.user = form.cleaned_data['user']
            instance.site = request.kbsite
            instance.save()
            messages.success(request, 'Token created.')
            return redirect('kegadmin-tokens')
    context['form'] = form
    return render_to_response('kegadmin/add_token.html', context_instance=context)


@staff_member_required
def beer_type_list(request):
    context = RequestContext(request)
    beers = models.BeerType.objects.all().order_by('name')
    paginator = Paginator(beers, 25)

    page = request.GET.get('page')
    try:
        beers = paginator.page(page)
    except PageNotAnInteger:
        beers = paginator.page(1)
    except EmptyPage:
        beers = paginator.page(paginator.num_pages)

    context['beers'] = beers
    return render_to_response('kegadmin/beer_type_list.html', context_instance=context)


@staff_member_required
def beer_type_detail(request, beer_id):
    btype = get_object_or_404(models.BeerType, id=beer_id)

    form = forms.BeerTypeForm(instance=btype)
    if request.method == 'POST':
        form = forms.BeerTypeForm(request.POST, instance=btype)
        if form.is_valid():
            btype = form.save()
            new_image = request.FILES.get('new_image')
            if new_image:
                pic = models.Picture.objects.create()
                pic.image.save(new_image.name, new_image)
                pic.save()
                btype.image = pic
                btype.save()

            messages.success(request, 'Beer type updated.')

    context = RequestContext(request)
    context['beer_type'] = btype
    context['form'] = form
    return render_to_response('kegadmin/beer_type_detail.html', context_instance=context)


@staff_member_required
def brewer_list(request):
    context = RequestContext(request)
    brewers = models.Brewer.objects.all().order_by('name')
    paginator = Paginator(brewers, 25)

    page = request.GET.get('page')
    try:
        brewers = paginator.page(page)
    except PageNotAnInteger:
        brewers = paginator.page(1)
    except EmptyPage:
        brewers = paginator.page(paginator.num_pages)

    context['brewers'] = brewers
    return render_to_response('kegadmin/brewer_list.html', context_instance=context)

@staff_member_required
def brewer_detail(request, brewer_id):
    brewer = get_object_or_404(models.Brewer, id=brewer_id)

    form = forms.BrewerForm(instance=brewer)
    if request.method == 'POST':
        form = forms.BrewerForm(request.POST, instance=brewer)
        if form.is_valid():
            form.save()
            messages.success(request, 'Brewer updated.')

    context = RequestContext(request)
    context['brewer'] = brewer
    context['form'] = form
    return render_to_response('kegadmin/brewer_detail.html', context_instance=context)


@staff_member_required
def beer_style_list(request):
    context = RequestContext(request)
    styles = models.BeerStyle.objects.all().order_by('name')
    paginator = Paginator(styles, 25)

    page = request.GET.get('page')
    try:
        styles = paginator.page(page)
    except PageNotAnInteger:
        styles = paginator.page(1)
    except EmptyPage:
        styles = paginator.page(paginator.num_pages)

    context['styles'] = styles
    return render_to_response('kegadmin/beer_style_list.html', context_instance=context)

@staff_member_required
def beer_style_detail(request, style_id):
    style = get_object_or_404(models.BeerStyle, id=style_id)

    form = forms.BeerStyleForm(instance=style)
    if request.method == 'POST':
        form = forms.BeerStyleForm(request.POST, instance=style)
        if form.is_valid():
            form.save()
            messages.success(request, 'Beer style updated.')

    context = RequestContext(request)
    context['style'] = style
    context['form'] = form
    return render_to_response('kegadmin/beer_style_detail.html', context_instance=context)

@staff_member_required
def autocomplete_beer_type(request):
    context = RequestContext(request)
    search = request.GET.get('q')
    if search:
        types = models.BeerType.objects.filter(Q(name__icontains=search) | Q(brewer__name__icontains=search))
    else:
        types = models.BeerType.objects.all()
    types = types[:10]  # autocomplete widget limited to 10
    values = []
    for beer in types:
        values.append({
          'name': beer.name,
          'id': beer.id,
          'brewer_name': beer.brewer.name,
          'brewer_id': beer.brewer.id,
          'style_name': beer.style.name,
          'style_id': beer.style.id,
      })
    return HttpResponse(kbjson.dumps(values, indent=None),
      mimetype='application/json', status=200)

@staff_member_required
def autocomplete_user(request):
    context = RequestContext(request)
    search = request.GET.get('q')
    if search:
        users = models.User.objects.filter(Q(username__icontains=search) | Q(email__icontains=search))
    else:
        users = models.User.objects.all()
    users = users[:10]  # autocomplete widget limited to 10
    values = []
    for user in users:
        values.append({
          'username': user.username,
          'id': user.id,
          'email': user.email,
          'is_active': user.is_active,
      })
    return HttpResponse(kbjson.dumps(values, indent=None),
      mimetype='application/json', status=200)

@staff_member_required
def autocomplete_token(request):
    context = RequestContext(request)
    search = request.GET.get('q')
    if search:
        tokens = models.AuthenticationToken.objects.filter(
            Q(token_value__icontains=search) | Q(nice_name__icontains=search))
    else:
        tokens = models.AuthenticationToken.objects.all()
    tokens = tokens[:10]  # autocomplete widget limited to 10
    values = []
    for token in tokens:
        values.append({
          'username': token.user.username,
          'id': token.id,
          'auth_device': token.auth_device,
          'token_value': token.token_value,
          'enabled': token.enabled,
      })
    return HttpResponse(kbjson.dumps(values, indent=None),
      mimetype='application/json', status=200)

@staff_member_required
def plugin_settings(request, plugin_name):
    context = RequestContext(request)
    plugin = request.plugins.get(plugin_name, None)
    if not plugin:
        raise Http404('Plugin "%s" not loaded' % plugin_name)

    view = plugin.get_admin_settings_view()
    if not view:
        raise Http404('No settings for this plugin')

    return view(request, plugin)

@staff_member_required
def logs(request):
    context = RequestContext(request)
    context['errors'] = logger.get_cached_logs()
    return render_to_response('kegadmin/logs.html', context_instance=context)



# Copyright 2012 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import sys

from pykeg import EPOCH

from django.core.management.base import NoArgsCommand
from django.db import connection

from django.contrib.staticfiles.management.commands import collectstatic
from south.management.commands import migrate
from south.management.commands import syncdb
from pykeg.core.management.commands import kb_regen_stats
from pykeg.core import models

def run(cmd, args=[]):
    cmdname = cmd.__module__.split('.')[-1]
    arg_str = ' '.join('%s' % a for a in args)
    print '--- Running command: %s %s' % (cmdname, arg_str)
    cmd.run_from_argv([sys.argv[0], cmdname] + args)

class Command(NoArgsCommand):
    help = u'Perform post-upgrade tasks.'

    def handle(self, **options):
        self.do_epoch_upgrades()
        run(syncdb.Command(), args=['--noinput', '-v', '0'])
        run(migrate.Command(), args=['-v', '0'])
        run(kb_regen_stats.Command())
        run(collectstatic.Command(), args=['--noinput'])

        site = models.KegbotSite.get()
        site.epoch = EPOCH
        site.save()

    def do_epoch_upgrades(self):
        installed = models.KegbotSite.get().epoch
        if installed == EPOCH or not installed:
            return

        print 'Performing epoch upgrades ...'
        for version in range(installed + 1, EPOCH + 1):
            fn = getattr(self, 'to_%s' % version, None)
            if fn:
                print '  ~ %s' % version
                fn()
            else:
                print '  ~ %s (no-op)' % version

    def to_101(self):
        cursor = connection.cursor()
        cursor.execute('DELETE FROM south_migrationhistory WHERE app_name = %s', ['twitter'])

    def to_102(self):
        cursor = connection.cursor()
        cursor.execute('DELETE FROM south_migrationhistory WHERE app_name = %s', ['foursquare'])

    def to_103(self):
        cursor = connection.cursor()
        cursor.execute('DELETE FROM south_migrationhistory WHERE app_name = %s', ['untappd'])



# Copyright 2012 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.conf import settings
from django.http import HttpResponse

from . import util

import logging
import sys

LOGGER = logging.getLogger(__name__)

class ApiRequestMiddleware:
    def process_view(self, request, view_func, view_args, view_kwargs):
        request.is_kb_api_request = util.is_api_request(request)
        if not request.is_kb_api_request:
            # Not an API request. Skip me.
            return None

        try:
            if request.need_setup:
                raise ValueError('Setup required')
            elif request.need_upgrade:
                raise ValueError('Upgrade required')

            need_auth = util.needs_auth(view_func)
            privacy = request.kbsite.settings.privacy
            if request.path in ('/api/login/', '/api/get-api-key/'):
                # API request to whitelisted path.
                need_auth = False
            else:
                # API request to non-whitelisted path, in non-public site privacy mode.
                # Demand API key.
                if privacy == 'members' and not request.user.is_authenticated():
                    need_auth = True
                elif privacy == 'staff' and not request.user.is_staff:
                    need_auth = True

            if need_auth:
                util.check_api_key(request)

            if request.method == 'GET':
                cached = request.kbcache.gen_get(cache_key(request))
                if cached:
                    response = util.build_response(request, cached, 200)
                    response.is_from_cache = True
                    return response

            return None

        except Exception, e:
            return util.wrap_exception(request, e)


class ApiResponseMiddleware:
    def process_exception(self, request, exception):
        """Wraps exceptions for API requests."""
        if util.is_api_request(request):
            return util.wrap_exception(request, exception)
        return None

    def process_response(self, request, response):
        if not util.is_api_request(request):
            return response

        if not isinstance(response, HttpResponse):
            data = util.prepare_data(response)
            data['meta'] = {
              'result': 'ok'
            }
            response = util.build_response(request, data, 200)

            if request.method == 'GET' and response.status_code == 200:
                if not getattr(response, 'is_from_cache', False):
                    request.kbcache.gen_set(cache_key(request), data)
        response['Cache-Control'] = 'max-age=0'
        return response


def cache_key(request):
    return 'api:%s' % request.get_full_path()



# Copyright 2014 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import logging
logger = logging.getLogger('notification')

from django.conf import settings
from django.utils.module_loading import import_by_path
from pykeg.notification import models
from pykeg.core import models as core_models

__all__ = ['get_backends', 'handle_new_system_events']

def get_backends():
    """Returns the enabled notification backend(s)."""
    backend_names = settings.NOTIFICATION_BACKENDS
    backends = [import_by_path(n)() for n in backend_names]
    return backends

def handle_new_system_events(events):
    """Processes newly-generated system events.

    Args:
        events: an iterable of models.SystemEvent objects

    Returns:
        None 
    """
    backends = get_backends()
    logger.info('Handling %s event(s)' % len(events))
    logger.info('Num backends: %s' % len(backends))
    for event in events:
        handle_single_event(event, backends)

def handle_single_event(event, backends):
    kind = event.kind
    logger.info('Processing event: %s' % event.kind)
    
    for backend in backends:
        backend_name = str(backend.__class__)
        prefs = models.NotificationSettings.objects.filter(backend=backend_name)
        SystemEvent = core_models.SystemEvent

        if kind == SystemEvent.KEG_TAPPED:
            prefs = prefs.filter(keg_tapped=True)
        elif kind == SystemEvent.SESSION_STARTED:
            prefs = prefs.filter(session_started=True)
        elif kind == SystemEvent.KEG_VOLUME_LOW:
            prefs = prefs.filter(keg_volume_low=True)
        elif kind == SystemEvent.KEG_ENDED:
            prefs = prefs.filter(keg_ended=True)
        else:
            logger.info('Unknown kind: %s' % kind)
            prefs = []

        logger.debug('Matching prefs: %s' % str(prefs))
        for matching_pref in prefs:
            user = matching_pref.user
            logger.info('Notifying %s for event %s' % (user, kind))
            backend.notify(event, user)




"""Defines HAVE_* flags for several optional packages."""

import imp
from django.core.exceptions import ImproperlyConfigured

try:
    import debug_toolbar
    HAVE_DEBUG_TOOLBAR = True
except ImportError:
    HAVE_DEBUG_TOOLBAR = False

try:
    import raven.contrib.django
    HAVE_RAVEN = True
except ImportError:
    HAVE_RAVEN = False

try:
    import django_statsd
    HAVE_STATSD = True
except ImproperlyConfigured:
    # Good enough; statsd is fussy on import.
    HAVE_STATSD = True
except ImportError:
    HAVE_STATSD = False

try:
    import storages
    HAVE_STORAGES = True
except ImportError:
    HAVE_STORAGES = False

try:
    import memcache
    HAVE_MEMCACHE = True
except ImportError:
    HAVE_MEMCACHE = False

try:
    import pylibmc
    HAVE_PYLIBMC = True
except ImportError:
    HAVE_PYLIBMC = False

try:
    import debug_toolbar_memcache
    HAVE_MEMCACHE_TOOLBAR = True
except ImportError:
    HAVE_MEMCACHE_TOOLBAR = False

try:
    import johnny
    HAVE_JOHNNY_CACHE = True
except ImportError:
    HAVE_JOHNNY_CACHE = False

try:
    import djcelery_email
    HAVE_CELERY_EMAIL = True
except ImportError:
    HAVE_CELERY_EMAIL = False



from django import forms
from django.contrib.auth.forms import AuthenticationForm

class LoginForm(AuthenticationForm):
    next_page = forms.CharField(required=False, widget=forms.HiddenInput)


class MugshotForm(forms.Form):
    new_mugshot = forms.ImageField(required=True)


class RegenerateApiKeyForm(forms.Form):
    pass



from django.conf.urls import include
from django.conf.urls import patterns
from django.conf.urls import url
from django.views.generic.base import TemplateView

from pykeg.web.registration.views import RegistrationView
from registration.backends.default.views import ActivationView


urlpatterns = patterns('',
                       url(r'^activate/complete/$',
                           TemplateView.as_view(template_name='registration/activation_complete.html'),
                           name='registration_activation_complete'),
                       # Activation keys get matched by \w+ instead of the more specific
                       # [a-fA-F0-9]{40} because a bad activation key should still get to the view;
                       # that way it can return a sensible "invalid key" message instead of a
                       # confusing 404.
                       url(r'^activate/(?P<activation_key>\w+)/$',
                           ActivationView.as_view(),
                           name='registration_activate'),
                       url(r'^register/$',
                           RegistrationView.as_view(),
                           name='registration_register'),
                       url(r'^register/complete/$',
                           TemplateView.as_view(template_name='registration/registration_complete.html'),
                           name='registration_complete'),
                       url(r'^register/closed/$',
                           TemplateView.as_view(template_name='registration/registration_closed.html'),
                           name='registration_disallowed'),
                       (r'', include('registration.auth_urls')),
                       )



# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.contrib import messages
from django.core.urlresolvers import reverse
from django.shortcuts import redirect
from socialregistration.clients.oauth import OAuthError
from socialregistration.contrib.foursquare.client import Foursquare
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.shortcuts import render_to_response
from django.template import RequestContext
import oauth2 as oauth

import foursquare

from . import forms

class FoursquareClient(Foursquare):
    def set_callback_url(self, url):
        self.callback_url = url

    def get_callback_url(self):
        return self.callback_url


@staff_member_required
def admin_settings(request, plugin):
    context = RequestContext(request)
    settings_form = plugin.get_site_settings_form()

    if request.method == 'POST':
        if 'submit-settings' in request.POST:
            settings_form = forms.SiteSettingsForm(request.POST)
            if settings_form.is_valid():
                client_id = settings_form.cleaned_data['client_id']
                client_secret = settings_form.cleaned_data['client_secret']
                venue_id = settings_form.cleaned_data['venue_id']
                client = foursquare.Foursquare(client_id=client_id, client_secret=client_secret)
                venue = None
                try:
                    # Search for the venue. Simultaneous validates all three credentials.
                    if venue_id:
                        venue = client.venues(venue_id)
                except foursquare.FoursquareException, e:
                    messages.error(request, 'Error testing Foursquare connection: %s' % str(e))
                else:
                    plugin.save_site_settings_form(settings_form)
                    plugin.save_venue_detail(venue)
                    messages.success(request, 'Settings updated')

    context['plugin'] = plugin
    context['settings_form'] = settings_form
    context['venue_detail'] = plugin.get_venue_detail()

    return render_to_response('contrib/foursquare/foursquare_admin_settings.html', context_instance=context)


@login_required
def user_settings(request, plugin):
    context = RequestContext(request)
    user = request.user

    settings_form = plugin.get_user_settings_form(user)

    if request.method == 'POST':
        if 'submit-settings' in request.POST:
            settings_form = forms.UserSettingsForm(request.POST)
            if settings_form.is_valid():
                plugin.save_user_settings_form(user, settings_form)
                messages.success(request, 'Settings updated')

    context['plugin'] = plugin
    context['venue'] = plugin.get_venue_detail()
    context['profile'] = plugin.get_user_profile(user)
    context['settings_form'] = settings_form

    return render_to_response('contrib/foursquare/foursquare_user_settings.html', context_instance=context)


@staff_member_required
def auth_redirect(request):
    if 'submit-remove' in request.POST:
        plugin = request.plugins.get('foursquare')
        plugin.save_user_profile(request.user, None)
        plugin.save_user_token(request.user, '')
        messages.success(request, 'Removed Foursquare account.')
        return redirect('account-plugin-settings', plugin_name='foursquare')

    plugin = request.plugins['foursquare']
    client = get_client(*plugin.get_credentials())

    url = request.kbsite.settings.reverse_full('plugin-foursquare-callback')
    client.set_callback_url(url)

    request.session['foursquare_client'] = client

    try:
        return redirect(client.get_redirect_url())
    except OAuthError, error:
        messages.error(request, 'Error: %s' % str(error))
        return redirect('account-plugin-settings', plugin_name='foursquare')


@staff_member_required
def auth_callback(request):
    try:
        client = request.session['foursquare_client']
        del request.session['foursquare_client']
        token = client.complete(dict(request.GET.items()))
    except KeyError:
        messages.error(request, 'Session expired.')
    except OAuthError, error:
        messages.error(request, str(error))
    else:
        plugin = request.plugins.get('foursquare')
        profile = client.get_user_info()
        token = client.get_access_token()
        plugin.save_user_profile(request.user, profile)
        plugin.save_user_token(request.user, token)

        username = '%s %s' % (profile.get('firstName'), profile.get('lastName'))
        messages.success(request, 'Successfully linked to foursquare user %s' % username)

    return redirect('account-plugin-settings', plugin_name='foursquare')


def get_client(client_id, client_secret):
    client = FoursquareClient()
    client.client_id = client_id
    client.secret = client_secret
    return client



# encoding: utf-8
import datetime
import sys
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    no_dry_run = True

    def forwards(self, orm):
        
        # Changing field 'Drink.user'
        db.alter_column('core_drink', 'user_id', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['auth.User'], null=True, blank=True))

        # Unfortunately, some systems had multiple users with the
        # '__default_user__' label.  Ask the user to give the correct id.
        guest_ids = {}
        for profile in orm.UserProfile.objects.all():
          if '__default_user__' in (l.labelname for l in profile.labels.all()):
            user = orm['auth.User'].objects.get(id=profile.user_id)
            guest_ids[profile.user_id] = user.username

        if len(guest_ids) > 1:
          print "Error: too many users with __default_user__ label"
          print "Please remove the label from all but the single 'guest' user."
          print ''
          for gid, username in guest_ids.iteritems():
            print '[%s]  %s' % (gid, username)
          print ''
          gid = None
          while gid is None:
            gid_str = raw_input('REAL guest user id (number): ')
            try:
              gid = int(gid_str.strip())
              break
            except ValueError:
              print '\nError, invalid id.'
        elif len(guest_ids) == 1:
          gid = guest_ids.keys()[0]
        else:
          # No guest user, nothing to do.
          return

        for d in orm.Drink.objects.all():
          if d.user_id == gid:
            print '  guest drink:', d
            d.user_id = None
            d.save()


    def backwards(self, orm):
        
        new_guest, _ = orm['auth.User'].objects.get_or_create(username='__unknown__')
        profile, _ = orm.UserProfile.objects.get_or_create(user=new_guest, 
            defaults={'gender':'male', 'weight':0})
        label = orm.UserLabel.objects.create(labelname='__default_user__')
        profile.labels.add(label.id)

        guest_id = new_guest.id

        # Changing field 'Drink.user'
        db.alter_column('core_drink', 'user_id', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['auth.User']))


    models = {
        'auth.group': {
            'Meta': {'object_name': 'Group'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        'auth.permission': {
            'Meta': {'unique_together': "(('content_type', 'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        'auth.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True', 'blank': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False', 'blank': 'True'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False', 'blank': 'True'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        'beerdb.beerimage': {
            'Meta': {'object_name': 'BeerImage'},
            'added': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'edited': ('django.db.models.fields.DateTimeField', [], {}),
            'id': ('django.db.models.fields.CharField', [], {'max_length': '36', 'primary_key': 'True'}),
            'num_views': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'original_image': ('django.db.models.fields.files.ImageField', [], {'max_length': '100'}),
            'revision': ('django.db.models.fields.IntegerField', [], {'default': '0'})
        },
        'beerdb.beerstyle': {
            'Meta': {'object_name': 'BeerStyle'},
            'added': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'edited': ('django.db.models.fields.DateTimeField', [], {}),
            'id': ('django.db.models.fields.CharField', [], {'max_length': '36', 'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'revision': ('django.db.models.fields.IntegerField', [], {'default': '0'})
        },
        'beerdb.beertype': {
            'Meta': {'object_name': 'BeerType'},
            'abv': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'added': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'brewer': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['beerdb.Brewer']"}),
            'calories_oz': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'carbs_oz': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'edited': ('django.db.models.fields.DateTimeField', [], {}),
            'edition': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.CharField', [], {'max_length': '36', 'primary_key': 'True'}),
            'image': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'beers'", 'null': 'True', 'to': "orm['beerdb.BeerImage']"}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'original_gravity': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'revision': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            'specific_gravity': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'style': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['beerdb.BeerStyle']"})
        },
        'beerdb.brewer': {
            'Meta': {'object_name': 'Brewer'},
            'added': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'country': ('pykeg.core.fields.CountryField', [], {'default': "'USA'", 'max_length': '3'}),
            'description': ('django.db.models.fields.TextField', [], {'default': "''", 'null': 'True', 'blank': 'True'}),
            'edited': ('django.db.models.fields.DateTimeField', [], {}),
            'id': ('django.db.models.fields.CharField', [], {'max_length': '36', 'primary_key': 'True'}),
            'image': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'brewers'", 'null': 'True', 'to': "orm['beerdb.BeerImage']"}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'origin_city': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'origin_state': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'production': ('django.db.models.fields.CharField', [], {'default': "'commercial'", 'max_length': '128'}),
            'revision': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            'url': ('django.db.models.fields.URLField', [], {'default': "''", 'max_length': '200', 'null': 'True', 'blank': 'True'})
        },
        'contenttypes.contenttype': {
            'Meta': {'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'core.authenticationtoken': {
            'Meta': {'unique_together': "(('auth_device', 'token_value'),)", 'object_name': 'AuthenticationToken'},
            'auth_device': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'enabled': ('django.db.models.fields.BooleanField', [], {'default': 'True', 'blank': 'True'}),
            'expires': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'pin': ('django.db.models.fields.CharField', [], {'max_length': '256', 'null': 'True', 'blank': 'True'}),
            'token_value': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']", 'null': 'True', 'blank': 'True'})
        },
        'core.bac': {
            'Meta': {'object_name': 'BAC'},
            'bac': ('django.db.models.fields.FloatField', [], {}),
            'drink': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['core.Drink']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'rectime': ('django.db.models.fields.DateTimeField', [], {}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']"})
        },
        'core.config': {
            'Meta': {'object_name': 'Config'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255'}),
            'value': ('django.db.models.fields.TextField', [], {})
        },
        'core.drink': {
            'Meta': {'object_name': 'Drink'},
            'endtime': ('django.db.models.fields.DateTimeField', [], {}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['core.Keg']", 'null': 'True', 'blank': 'True'}),
            'starttime': ('django.db.models.fields.DateTimeField', [], {}),
            'status': ('django.db.models.fields.CharField', [], {'default': "'valid'", 'max_length': '128'}),
            'ticks': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']", 'null': 'True', 'blank': 'True'}),
            'volume_ml': ('django.db.models.fields.FloatField', [], {})
        },
        'core.drinkingsession': {
            'Meta': {'object_name': 'DrinkingSession'},
            'drinks': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['core.Drink']", 'symmetrical': 'False'}),
            'endtime': ('django.db.models.fields.DateTimeField', [], {}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'kegs': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['core.Keg']", 'symmetrical': 'False'}),
            'starttime': ('django.db.models.fields.DateTimeField', [], {}),
            'users': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.User']", 'symmetrical': 'False'})
        },
        'core.drinkingsessionuserpart': {
            'Meta': {'unique_together': "(('session', 'user'),)", 'object_name': 'DrinkingSessionUserPart'},
            'endtime': ('django.db.models.fields.DateTimeField', [], {}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'user_parts'", 'to': "orm['core.DrinkingSession']"}),
            'starttime': ('django.db.models.fields.DateTimeField', [], {}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'session_parts'", 'to': "orm['auth.User']"}),
            'volume_ml': ('django.db.models.fields.FloatField', [], {'default': '0'})
        },
        'core.keg': {
            'Meta': {'object_name': 'Keg'},
            'description': ('django.db.models.fields.CharField', [], {'max_length': '256', 'null': 'True', 'blank': 'True'}),
            'enddate': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'origcost': ('django.db.models.fields.FloatField', [], {'default': '0', 'null': 'True', 'blank': 'True'}),
            'size': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['core.KegSize']"}),
            'startdate': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'status': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['beerdb.BeerType']"})
        },
        'core.kegsize': {
            'Meta': {'object_name': 'KegSize'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'volume_ml': ('django.db.models.fields.FloatField', [], {})
        },
        'core.kegstats': {
            'Meta': {'object_name': 'KegStats'},
            'date': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'stats'", 'unique': 'True', 'to': "orm['core.Keg']"}),
            'revision': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'stats': ('pykeg.core.jsonfield.JSONField', [], {'default': "'{}'"})
        },
        'core.kegtap': {
            'Meta': {'object_name': 'KegTap'},
            'current_keg': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['core.Keg']", 'null': 'True', 'blank': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'max_tick_delta': ('django.db.models.fields.PositiveIntegerField', [], {'default': '100'}),
            'meter_name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'ml_per_tick': ('django.db.models.fields.FloatField', [], {'default': '0.45454545454545453'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'temperature_sensor': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['core.ThermoSensor']", 'null': 'True', 'blank': 'True'})
        },
        'core.relaylog': {
            'Meta': {'object_name': 'RelayLog'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'status': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            'time': ('django.db.models.fields.DateTimeField', [], {})
        },
        'core.thermolog': {
            'Meta': {'object_name': 'Thermolog'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'sensor': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['core.ThermoSensor']"}),
            'temp': ('django.db.models.fields.FloatField', [], {}),
            'time': ('django.db.models.fields.DateTimeField', [], {})
        },
        'core.thermosensor': {
            'Meta': {'object_name': 'ThermoSensor'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'nice_name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'raw_name': ('django.db.models.fields.CharField', [], {'max_length': '256'})
        },
        'core.thermosummarylog': {
            'Meta': {'object_name': 'ThermoSummaryLog'},
            'date': ('django.db.models.fields.DateTimeField', [], {}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'max_temp': ('django.db.models.fields.FloatField', [], {}),
            'mean_temp': ('django.db.models.fields.FloatField', [], {}),
            'min_temp': ('django.db.models.fields.FloatField', [], {}),
            'num_readings': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'period': ('django.db.models.fields.CharField', [], {'default': "'daily'", 'max_length': '64'}),
            'sensor': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['core.ThermoSensor']"})
        },
        'core.userlabel': {
            'Meta': {'object_name': 'UserLabel'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'labelname': ('django.db.models.fields.CharField', [], {'max_length': '128'})
        },
        'core.userpicture': {
            'Meta': {'object_name': 'UserPicture'},
            'active': ('django.db.models.fields.BooleanField', [], {'default': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'image': ('django.db.models.fields.files.ImageField', [], {'max_length': '100'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']"})
        },
        'core.userprofile': {
            'Meta': {'object_name': 'UserProfile'},
            'gender': ('django.db.models.fields.CharField', [], {'max_length': '8'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'labels': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['core.UserLabel']", 'symmetrical': 'False'}),
            'mugshot': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['core.UserPicture']", 'null': 'True', 'blank': 'True'}),
            'user': ('django.db.models.fields.related.OneToOneField', [], {'to': "orm['auth.User']", 'unique': 'True'}),
            'weight': ('django.db.models.fields.FloatField', [], {})
        },
        'core.userstats': {
            'Meta': {'object_name': 'UserStats'},
            'date': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'revision': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'stats': ('pykeg.core.jsonfield.JSONField', [], {'default': "'{}'"}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'stats'", 'unique': 'True', 'to': "orm['auth.User']"})
        }
    }

    complete_apps = ['core']



# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Methods to generate cached statistics from drinks."""

import copy
import inspect
import itertools
import logging

from django.db import transaction

from pykeg.core import models
from kegbot.util import util

STAT_MAP = {}

def stat(statname):
    def decorate(f):
        setattr(f, 'statname', statname)
        return f
    return decorate

class StatsBuilder:
    def __init__(self, drink, drink_qs, previous=None):
        self.drink = drink
        self.drinks = drink_qs
        self.previous = util.AttrDict(copy.deepcopy(previous))
        self.STAT_MAP = {}
        for name, fn in inspect.getmembers(self, inspect.ismethod):
            if hasattr(fn, 'statname'):
                self.STAT_MAP[fn.statname] = fn

    def build(self, tag=None):
        stats = util.AttrDict()
        if not self.drink:
            return stats
        if self.previous:
            stats = copy.deepcopy(self.previous)
        for statname, fn in self.STAT_MAP.iteritems():
            previous = self.previous.get(statname, None)
            val = fn(previous)
            if val is None:
                raise ValueError('Stat generator for %s returned None' % statname)
            stats[statname] = val
        return stats

    @stat('last_drink_id')
    def LastDrinkId(self, previous):
        return self.drink.id

    @stat('total_volume_ml')
    def TotalVolume(self, previous):
        if not previous:
            return sum(drink.volume_ml for drink in self.drinks)
        else:
            return previous + self.drink.volume_ml

    @stat('total_pours')
    def TotalPours(self, previous):
        if not previous:
            return self.drinks.count()
        else:
            return previous + 1

    @stat('average_volume_ml')
    def AverageVolume(self, previous):
        if not previous:
            count = self.drinks.count()
            average = 0.0
            if count:
                average = sum(drink.volume_ml for drink in self.drinks) / float(count)
            return average
        else:
            vol = self.previous.total_volume_ml
            count = self.previous.total_pours
            vol += self.drink.volume_ml
            count += 1
            return vol / float(count)

    @stat('greatest_volume_ml')
    def GreatestVolume(self, previous):
        if not previous:
            res = 0
            drinks = self.drinks.order_by('-volume_ml')
            if drinks.count():
                res = drinks[0].volume_ml
            return float(res)
        else:
            return float(max(self.drink.volume_ml, previous))

    @stat('greatest_volume_id')
    def GreatestVolumeId(self, previous):
        if not previous:
            res = 0
            drinks = self.drinks.order_by('-volume_ml')
            if drinks.count():
                res = drinks[0].id
            return res
        else:
            if self.drink.volume_ml > self.previous.greatest_volume_ml:
                return self.drink.id
            return previous

    @stat('volume_by_day_of_week')
    def VolumeByDayOfweek(self, previous):
        if not previous:
            # Note: uses the session's start_time, rather than the drink's. This
            # causes late-night sessions to be reported for the day on which they were
            # started.
            volmap = {}
            for drink in self.drinks:
                weekday = str(drink.session.start_time.strftime('%w'))
                if weekday not in volmap:
                    volmap[weekday] = 0.0
                volmap[weekday] += drink.volume_ml
            return volmap
        else:
            drink_weekday = str(self.drink.session.start_time.strftime('%w'))
            previous[drink_weekday] = previous.get(drink_weekday, 0) + self.drink.volume_ml
            return previous

    @stat('registered_drinkers')
    def RegisteredDrinkers(self, previous):
        if previous is None:
            drinkers = set()
            for drink in self.drinks:
                if drink.user:
                    drinkers.add(str(drink.user.username))
            return list(drinkers)
        else:
            if self.drink.user:
                username = str(self.drink.user.username)
                if username not in previous:
                    previous.append(str(username))
            return previous

    @stat('sessions_count')
    def SessionsCount(self, previous):
        if not previous:
            all_sessions = set()
            for drink in self.drinks:
                all_sessions.add(drink.session.id)
            return len(all_sessions)
        else:
            first_drink = self.drink.session.drinks.order_by('id')[0]
            if self.drink.id == first_drink.id:
                previous += 1
            return previous

    @stat('volume_by_year')
    def VolumeByYear(self, previous):
        if not previous:
            ret = {}
            for drink in self.drinks:
                year = str(drink.time.year)
                ret[year] = ret.get(year, 0) + drink.volume_ml
            return ret
        else:
            year = str(self.drink.time.year)
            previous[year] = previous.get(year, 0) + self.drink.volume_ml
            return previous

    @stat('has_guest_pour')
    def HasGuestPour(self, previous):
        if not previous:
            return self.drinks.filter(user_id=None).count() > 0
        else:
            return bool(previous or self.drink.user is None)

    @stat('volume_by_drinker')
    def VolumeByDrinker(self, previous):
        if not previous:
            volmap = {}
            for drink in self.drinks:
                if drink.user:
                    u = str(drink.user.username)
                else:
                    u = ''
                volmap[u] = float(volmap.get(u, 0) + drink.volume_ml)
            return volmap
        else:
            if self.drink.user:
                u = str(self.drink.user.username)
            else:
                u = ''
            orig = previous.get(u, 0)
            previous[u] = float(orig + self.drink.volume_ml)
            return previous


def invalidate(drink):
    """Clears all statistics.

    Args:
        drink: The starting point to invalidate; all stats starting from this
            drink will be deleted.  If None, deletes *all* stats.
    """
    with transaction.atomic():
        if drink:
            models.SystemStats.objects.filter(drink_id__gte=drink.id).delete()
            models.KegStats.objects.filter(drink_id__gte=drink.id).delete()
            models.UserStats.objects.filter(drink_id__gte=drink.id).delete()
            models.SessionStats.objects.filter(drink_id__gte=drink.id).delete()
        else:
            models.SystemStats.objects.all().delete()
            models.KegStats.objects.all().delete()
            models.UserStats.objects.all().delete()
            models.SessionStats.objects.all().delete()

def _get_previous(drink, qs):
    qs = qs.filter(drink_id__lt=drink.id).order_by('-id')[:1]
    if qs.count():
        return qs[0].stats
    return {}

def build_system_stats(drink):
    """Builds (but does not save) system stats dictionary for drink."""
    qs = models.Drink.objects.filter(id__lte=drink.id)
    previous = _get_previous(drink, models.SystemStats.objects.all())
    builder = StatsBuilder(drink, qs, previous)
    return builder.build('system')

def generate_system_stats(drink):
    """Builds and saves system stats record for drink."""
    stats = build_system_stats(drink)
    return models.SystemStats.objects.create(drink=drink, stats=stats)

def build_keg_stats(drink):
    """Builds (but does not save) keg stats dictionary for drink."""
    if drink.keg:
        qs = models.Drink.objects.filter(id__lte=drink.id, keg_id=drink.keg.id)
        previous = _get_previous(drink,
            models.KegStats.objects.filter(keg=drink.keg))
        builder = StatsBuilder(drink, qs, previous)
        return builder.build('keg ' + str(drink.keg.id))

def generate_keg_stats(drink):
    """Builds and saves keg stats record for drink."""
    stats = build_keg_stats(drink)
    if stats:
        return models.KegStats.objects.create(drink=drink, keg=drink.keg, stats=stats)

def build_user_stats(drink):
    """Builds (but does not save) user stats dictionary for drink."""
    if drink.user:
        user_id = drink.user.id
    else:
        user_id = None
    qs = models.Drink.objects.filter(id__lte=drink.id, user_id=user_id)
    previous = _get_previous(drink,
        models.UserStats.objects.filter(user=drink.user))
    builder = StatsBuilder(drink, qs, previous)
    return builder.build('user ' + str(user_id))

def generate_user_stats(drink):
    """Builds and saves user stats record for drink."""
    stats = build_user_stats(drink)
    return models.UserStats.objects.create(drink=drink, user=drink.user, stats=stats)

def build_session_stats(drink):
    """Builds (but does not save) session stats dictionary for drink."""
    if drink.session:
        qs = models.Drink.objects.filter(id__lte=drink.id, session_id=drink.session.id)
        previous = _get_previous(drink,
            models.SessionStats.objects.filter(session=drink.session))
        builder = StatsBuilder(drink, qs, previous)
        return builder.build('session ' + str(drink.session.id))

def generate_session_stats(drink):
    """Builds and saves session stats record for drink."""
    stats = build_session_stats(drink)
    if stats:
        return models.SessionStats.objects.create(drink=drink,
            session=drink.session, stats=stats)

def generate(drink, invalidate_first=True):
    """Generate all stats for this drink.

    Args:
        drink: The drink.
        invalidate_first: If True, deletes any existing records for (or after)
            this drink first.
    """
    with transaction.atomic():
        if invalidate_first:
            invalidate(drink)

        generate_system_stats(drink)
        generate_keg_stats(drink)
        generate_user_stats(drink)
        generate_session_stats(drink)

if __name__ == '__main__':
    import cProfile
    command = """main()"""
    cProfile.runctx( command, globals(), locals(), filename="stats.profile" )
